// Kernels for MD5 - vectorized for AMD...

// Things we should define in the calling code...
//#define VECTOR_WIDTH 2
//#define CPU_DEBUG 1
//#define BITALIGN
//#define NVIDIA_HACKS 1
//#define PASSWORD_LENGTH 6


#if VECTOR_WIDTH == 1
    #define vector_type uint
    #define vload_type vload1
    #define vstore_type vstore1
    #define grt_vector_1 1
#elif VECTOR_WIDTH == 2
    #define vector_type uint2
    #define vload_type vload2
    #define vstore_type vstore2
    #define grt_vector_2 1
#elif VECTOR_WIDTH == 4
    #define vector_type uint4
    #define vload_type vload4
    #define vstore_type vstore4
    #define grt_vector_4 1
#elif VECTOR_WIDTH == 8
ERROR NOT SUPPORTED
    #define vector_type uint8
    #define vload_type vload8
    #define vstore_type vstore8
    #define grt_vector_8 1
#elif VECTOR_WIDTH == 16
ERROR NOT SUPPORTED
    #define vector_type uint16
    #define vload_type vload16
    #define vstore_type vstore16
    #define grt_vector_16 1
#endif


#ifdef CPU_DEBUG
#pragma OPENCL EXTENSION cl_amd_printf : enable
#endif


// Hash defines

#define MD5S11 7
#define MD5S12 12
#define MD5S13 17
#define MD5S14 22
#define MD5S21 5
#define MD5S22 9
#define MD5S23 14
#define MD5S24 20
#define MD5S31 4
#define MD5S32 11
#define MD5S33 16
#define MD5S34 23
#define MD5S41 6
#define MD5S42 10
#define MD5S43 15
#define MD5S44 21

// OPTIMIZED MD5 FUNCTIONS HERE
//#define MD5F(x,y,z) (((y ^ z) & x) ^ z)
//#define MD5G(x,y,z) (((x & y) & z) ^ y)

/* F, G, H and I are basic MD5 functions.
 */
#define MD5F(x, y, z) (((x) & (y)) | ((~x) & (z)))
#define MD5G(x, y, z) (((x) & (z)) | ((y) & (~z)))
#define MD5H(x, y, z) ((x) ^ (y) ^ (z))
#define MD5I(x, y, z) ((y) ^ ((x) | (~z)))


/* ROTATE_LEFT rotates x left n bits.
 */

#ifdef BITALIGN
#pragma OPENCL EXTENSION cl_amd_media_ops : enable
#define MD5ROTATE_LEFT(x, y) amd_bitalign(x, x, (uint)(32 - y))
#define MD5FF(a, b, c, d, x, s, ac) { \
 (a) += amd_bytealign((b),(c),(d)) + (x) + (vector_type)(ac); \
 (a) = MD5ROTATE_LEFT ((a), (s)); \
 (a) += (b); \
  }

#elif NVIDIA_HACKS
#define MD5ROTATE_LEFT(x, n) (((x) << (n)) | ((x) >> (32-(n))))
#define MD5FF(a, b, c, d, x, s, ac) { \
 (a) += MD5F ((b), (c), (d)) + (x) + (vector_type)(ac); \
 (a) = MD5ROTATE_LEFT ((a), (s)); \
 (a) += (b); \
  }
#else
#define MD5ROTATE_LEFT(x, y) rotate((vector_type)x, (uint)y)
#define MD5FF(a, b, c, d, x, s, ac) { \
 (a) += MD5F ((b), (c), (d)) + (x) + (vector_type)(ac); \
 (a) = MD5ROTATE_LEFT ((a), (s)); \
 (a) += (b); \
  }
#endif


//#define MD5ROTATE_LEFT(x, n) (((x) << (n)) | ((x) >> (32-(n))))

/* FF, GG, HH, and II transformations for rounds 1, 2, 3, and 4.
Rotation is separate from addition to prevent recomputation.
 */

/*#define MD5FF(a, b, c, d, x, s, ac) { \
 (a) += MD5F ((b), (c), (d)) + (x) + (vector_type)(ac); \
 (a) = MD5ROTATE_LEFT ((a), (s)); \
 (a) += (b); \
  }*/
/*
#pragma OPENCL EXTENSION cl_amd_media_ops : enable

#define MD5FF(a, b, c, d, x, s, ac) { \
 (a) += amd_bytealign((b),(c),(d)) + (x) + (vector_type)(ac); \
 (a) = MD5ROTATE_LEFT ((a), (s)); \
 (a) += (b); \
  }
*/
#define MD5GG(a, b, c, d, x, s, ac) { \
 (a) += MD5G ((b), (c), (d)) + (x) + (vector_type)(ac); \
 (a) = MD5ROTATE_LEFT ((a), (s)); \
 (a) += (b); \
  }
#define MD5HH(a, b, c, d, x, s, ac) { \
 (a) += MD5H ((b), (c), (d)) + (x) + (vector_type)(ac); \
 (a) = MD5ROTATE_LEFT ((a), (s)); \
 (a) += (b); \
  }
#define MD5II(a, b, c, d, x, s, ac) { \
 (a) += MD5I ((b), (c), (d)) + (x) + (vector_type)(ac); \
 (a) = MD5ROTATE_LEFT ((a), (s)); \
 (a) += (b); \
  }


inline void OpenCL_MD5(vector_type b0, vector_type b1, vector_type b2, vector_type b3, vector_type b4, vector_type b5, vector_type b6, vector_type b7,
			   vector_type b8, vector_type b9, vector_type b10, vector_type b11, vector_type b12, vector_type b13, vector_type b14, vector_type b15,
			   vector_type *a, vector_type *b, vector_type *c, vector_type *d) {
  // Load these with vectors.
  *a = (vector_type)0x67452301;
  *b = (vector_type)0xefcdab89;
  *c = (vector_type)0x98badcfe;
  *d = (vector_type)0x10325476;

  MD5FF (*a, *b, *c, *d, b0, MD5S11, 0xd76aa478); /* 1 */
  MD5FF (*d, *a, *b, *c, b1, MD5S12, 0xe8c7b756); /* 2 */
  MD5FF (*c, *d, *a, *b, b2, MD5S13, 0x242070db); /* 3 */
  MD5FF (*b, *c, *d, *a, b3, MD5S14, 0xc1bdceee); /* 4 */
  MD5FF (*a, *b, *c, *d, b4, MD5S11, 0xf57c0faf); /* 5 */
  MD5FF (*d, *a, *b, *c, b5, MD5S12, 0x4787c62a); /* 6 */
  MD5FF (*c, *d, *a, *b, b6, MD5S13, 0xa8304613); /* 7 */
  MD5FF (*b, *c, *d, *a, b7, MD5S14, 0xfd469501); /* 8 */
  MD5FF (*a, *b, *c, *d, b8, MD5S11, 0x698098d8); /* 9 */
  MD5FF (*d, *a, *b, *c, b9, MD5S12, 0x8b44f7af); /* 10 */
  MD5FF (*c, *d, *a, *b, b10, MD5S13, 0xffff5bb1); /* 11 */
  MD5FF (*b, *c, *d, *a, b11, MD5S14, 0x895cd7be); /* 12 */
  MD5FF (*a, *b, *c, *d, b12, MD5S11, 0x6b901122); /* 13 */
  MD5FF (*d, *a, *b, *c, b13, MD5S12, 0xfd987193); /* 14 */
  MD5FF (*c, *d, *a, *b, b14, MD5S13, 0xa679438e); /* 15 */
  MD5FF (*b, *c, *d, *a, b15, MD5S14, 0x49b40821); /* 16 */

 /* Round 2 */
  MD5GG (*a, *b, *c, *d, b1, MD5S21, 0xf61e2562); /* 17 */
  MD5GG (*d, *a, *b, *c, b6, MD5S22, 0xc040b340); /* 18 */
  MD5GG (*c, *d, *a, *b, b11, MD5S23, 0x265e5a51); /* 19 */
  MD5GG (*b, *c, *d, *a, b0, MD5S24, 0xe9b6c7aa); /* 20 */
  MD5GG (*a, *b, *c, *d, b5, MD5S21, 0xd62f105d); /* 21 */
  MD5GG (*d, *a, *b, *c, b10, MD5S22,  0x2441453); /* 22 */
  MD5GG (*c, *d, *a, *b, b15, MD5S23, 0xd8a1e681); /* 23 */
  MD5GG (*b, *c, *d, *a, b4, MD5S24, 0xe7d3fbc8); /* 24 */
  MD5GG (*a, *b, *c, *d, b9, MD5S21, 0x21e1cde6); /* 25 */
  MD5GG (*d, *a, *b, *c, b14, MD5S22, 0xc33707d6); /* 26 */
  MD5GG (*c, *d, *a, *b, b3, MD5S23, 0xf4d50d87); /* 27 */
  MD5GG (*b, *c, *d, *a, b8, MD5S24, 0x455a14ed); /* 28 */
  MD5GG (*a, *b, *c, *d, b13, MD5S21, 0xa9e3e905); /* 29 */
  MD5GG (*d, *a, *b, *c, b2, MD5S22, 0xfcefa3f8); /* 30 */
  MD5GG (*c, *d, *a, *b, b7, MD5S23, 0x676f02d9); /* 31 */
  MD5GG (*b, *c, *d, *a, b12, MD5S24, 0x8d2a4c8a); /* 32 */

  /* Round 3 */
  MD5HH (*a, *b, *c, *d, b5, MD5S31, 0xfffa3942); /* 33 */
  MD5HH (*d, *a, *b, *c, b8, MD5S32, 0x8771f681); /* 34 */
  MD5HH (*c, *d, *a, *b, b11, MD5S33, 0x6d9d6122); /* 35 */
  MD5HH (*b, *c, *d, *a, b14, MD5S34, 0xfde5380c); /* 36 */
  MD5HH (*a, *b, *c, *d, b1, MD5S31, 0xa4beea44); /* 37 */
  MD5HH (*d, *a, *b, *c, b4, MD5S32, 0x4bdecfa9); /* 38 */
  MD5HH (*c, *d, *a, *b, b7, MD5S33, 0xf6bb4b60); /* 39 */
  MD5HH (*b, *c, *d, *a, b10, MD5S34, 0xbebfbc70); /* 40 */
  MD5HH (*a, *b, *c, *d, b13, MD5S31, 0x289b7ec6); /* 41 */
  MD5HH (*d, *a, *b, *c, b0, MD5S32, 0xeaa127fa); /* 42 */
  MD5HH (*c, *d, *a, *b, b3, MD5S33, 0xd4ef3085); /* 43 */
  MD5HH (*b, *c, *d, *a, b6, MD5S34,  0x4881d05); /* 44 */
  MD5HH (*a, *b, *c, *d, b9, MD5S31, 0xd9d4d039); /* 45 */
  MD5HH (*d, *a, *b, *c, b12, MD5S32, 0xe6db99e5); /* 46 */
  MD5HH (*c, *d, *a, *b, b15, MD5S33, 0x1fa27cf8); /* 47 */
  MD5HH (*b, *c, *d, *a, b2, MD5S34, 0xc4ac5665); /* 48 */

  /* Round 4 */
  MD5II (*a, *b, *c, *d, b0, MD5S41, 0xf4292244); /* 49 */
  MD5II (*d, *a, *b, *c, b7, MD5S42, 0x432aff97); /* 50 */
  MD5II (*c, *d, *a, *b, b14, MD5S43, 0xab9423a7); /* 51 */
  MD5II (*b, *c, *d, *a, b5, MD5S44, 0xfc93a039); /* 52 */
  MD5II (*a, *b, *c, *d, b12, MD5S41, 0x655b59c3); /* 53 */
  MD5II (*d, *a, *b, *c, b3, MD5S42, 0x8f0ccc92); /* 54 */
  MD5II (*c, *d, *a, *b, b10, MD5S43, 0xffeff47d); /* 55 */
  MD5II (*b, *c, *d, *a, b1, MD5S44, 0x85845dd1); /* 56 */
  MD5II (*a, *b, *c, *d, b8, MD5S41, 0x6fa87e4f); /* 57 */
  MD5II (*d, *a, *b, *c, b15, MD5S42, 0xfe2ce6e0); /* 58 */
  MD5II (*c, *d, *a, *b, b6, MD5S43, 0xa3014314); /* 59 */
  MD5II (*b, *c, *d, *a, b13, MD5S44, 0x4e0811a1); /* 60 */
  MD5II (*a, *b, *c, *d, b4, MD5S41, 0xf7537e82); /* 61 */
  MD5II (*d, *a, *b, *c, b11, MD5S42, 0xbd3af235); /* 62 */
  MD5II (*c, *d, *a, *b, b2, MD5S43, 0x2ad7d2bb); /* 63 */
  MD5II (*b, *c, *d, *a, b9, MD5S44, 0xeb86d391); /* 64 */

  // Finally, add initial values, as this is the only pass we make.
  *a += (vector_type)0x67452301;
  *b += (vector_type)0xefcdab89;
  *c += (vector_type)0x98badcfe;
  *d += (vector_type)0x10325476;
}


inline void copySingleCharsetToShared(__local unsigned char *sharedCharset,
        __constant unsigned char *constantCharset) {

    // 32-bit accesses are used to help coalesce memory accesses.
    uint a;

    for (a = 0; a < (512); a++) {
        sharedCharset[a] = constantCharset[a];
    }
}


inline void padMDHash(int length,
                vector_type *b0, vector_type *b1, vector_type *b2, vector_type *b3, vector_type *b4, vector_type *b5, vector_type *b6, vector_type *b7,
		vector_type *b8, vector_type *b9, vector_type *b10, vector_type *b11, vector_type *b12, vector_type *b13, vector_type *b14, vector_type *b15) {

  // Set length properly (length in bits)
  *b14 = (vector_type)(length * 8);

  if (length == 0) {
    *b0 |= (vector_type)0x00000080;
    return;
  }
  if (length == 1) {
    *b0 |= (vector_type)0x00008000;
    return;
  }
  if (length == 2) {
    *b0 |= (vector_type)0x00800000;
    return;
  }
  if (length == 3) {
    *b0 |= (vector_type)0x80000000;
    return;
  }
  if (length == 4) {
    *b1 |= (vector_type)0x00000080;
    return;
  }
  if (length == 5) {
    *b1 |= (vector_type)0x00008000;
    return;
  }
  if (length == 6) {
    *b1 |= (vector_type)0x00800000;
    return;
  }
  if (length == 7) {
    *b1 |= (vector_type)0x80000000;
    return;
  }
  if (length == 8) {
    *b2 |= (vector_type)0x00000080;
    return;
  }
  if (length == 9) {
    *b2 |= (vector_type)0x00008000;
    return;
  }
  if (length == 10) {
    *b2 |= (vector_type)0x00800000;
    return;
  }
  if (length == 11) {
   *b2 |= (vector_type)0x80000000;
    return;
  }
  if (length == 12) {
    *b3 |= (vector_type)0x00000080;
    return;
  }
  if (length == 13) {
    *b3 |= (vector_type)0x00008000;
    return;
  }
  if (length == 14) {
    *b3 |= (vector_type)0x00800000;
    return;
  }
  if (length == 15) {
    *b3 |= (vector_type)0x80000000;
    return;
  }
  if (length == 16) {
    *b4 |= (vector_type)0x00000080;
    return;
  }
  if (length == 17) {
    *b4 |= (vector_type)0x00008000;
    return;
  }
  if (length == 18) {
    *b4 |= (vector_type)0x00800000;
    return;
  }
  if (length == 19) {
    *b4 |= (vector_type)0x80000000;
    return;
  }
  if (length == 20) {
    *b5 |= (vector_type)0x00000080;
    return;
  }
  if (length == 21) {
    *b5 |= (vector_type)0x00008000;
    return;
  }
  if (length == 22) {
    *b5 |= (vector_type)0x00800000;
    return;
  }
  if (length == 23) {
    *b5 |= (vector_type)0x80000000;
    return;
  }
  if (length == 24) {
    *b6 |= (vector_type)0x00000080;
    return;
  }
  if (length == 25) {
    *b6 |= (vector_type)0x00008000;
    return;
  }
  if (length == 26) {
    *b6 |= (vector_type)0x00800000;
    return;
  }
  if (length == 27) {
    *b6 |= (vector_type)0x80000000;
    return;
  }
  if (length == 28) {
    *b7 |= (vector_type)0x00000080;
    return;
  }
  if (length == 29) {
    *b7 |= (vector_type)0x00008000;
    return;
  }
  if (length == 30) {
    *b7 |= (vector_type)0x00800000;
    return;
  }
  if (length == 31) {
    *b7 |= (vector_type)0x80000000;
    return;
  }

}


inline void reduceSingleCharsetNormal(vector_type *b0, vector_type *b1, vector_type *b2,
        vector_type a, vector_type b, vector_type c, vector_type d,
        uint CurrentStep, __local unsigned char *charset, uint charset_offset, int PasswordLength, uint Device_Table_Index) {

    vector_type z;
    // Reduce it
    // First 3

    z = (a+(vector_type)CurrentStep+(vector_type)Device_Table_Index) % (vector_type)(256*256*256);
#if CPU_DEBUG
    printf("z.s0: %08x\n", z.s0);
#endif
#if grt_vector_1 || grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b0).s0 = (uint)charset[(z.s0 % 256) + charset_offset];
#endif
#if grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b0).s1 = (uint)charset[(z.s1 % 256) + charset_offset];
#endif
#if grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b0).s2 = (uint)charset[(z.s2 % 256) + charset_offset];
    (*b0).s3 = (uint)charset[(z.s3 % 256) + charset_offset];
#endif
#if grt_vector_8 || grt_vector_16
    (*b0).s4 = (uint)charset[(z.s4 % 256) + charset_offset];
    (*b0).s5 = (uint)charset[(z.s5 % 256) + charset_offset];
    (*b0).s6 = (uint)charset[(z.s6 % 256) + charset_offset];
    (*b0).s7 = (uint)charset[(z.s7 % 256) + charset_offset];
#endif
    if (PasswordLength == 1) {return;}

    z /= (vector_type)256;
#if grt_vector_1 || grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b0).s0 |= (uint)charset[(z.s0 % 256) + charset_offset] << 8;
#endif
#if grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b0).s1 |= (uint)charset[(z.s1 % 256) + charset_offset] << 8;
#endif
#if grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b0).s2 |= (uint)charset[(z.s2 % 256) + charset_offset] << 8;
    (*b0).s3 |= (uint)charset[(z.s3 % 256) + charset_offset] << 8;
#endif
#if grt_vector_8 || grt_vector_16
    (*b0).s4 |= (uint)charset[(z.s4 % 256) + charset_offset] << 8;
    (*b0).s5 |= (uint)charset[(z.s5 % 256) + charset_offset] << 8;
    (*b0).s6 |= (uint)charset[(z.s6 % 256) + charset_offset] << 8;
    (*b0).s7 |= (uint)charset[(z.s7 % 256) + charset_offset] << 8;
#endif
    if (PasswordLength == 2) {return;}

    z /= (vector_type)256;
#if grt_vector_1 || grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b0).s0 |= (uint)charset[(z.s0 % 256) + charset_offset] << 16;
#endif
#if grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b0).s1 |= (uint)charset[(z.s1 % 256) + charset_offset] << 16;
#endif
#if grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b0).s2 |= (uint)charset[(z.s2 % 256) + charset_offset] << 16;
    (*b0).s3 |= (uint)charset[(z.s3 % 256) + charset_offset] << 16;
#endif
#if grt_vector_8 || grt_vector_16
    (*b0).s4 |= (uint)charset[(z.s4 % 256) + charset_offset] << 16;
    (*b0).s5 |= (uint)charset[(z.s5 % 256) + charset_offset] << 16;
    (*b0).s6 |= (uint)charset[(z.s6 % 256) + charset_offset] << 16;
    (*b0).s7 |= (uint)charset[(z.s7 % 256) + charset_offset] << 16;
#endif
    if (PasswordLength == 3) {return;}

    // Second 3
    z = (b+(vector_type)CurrentStep+(vector_type)Device_Table_Index) % (vector_type)(256*256*256);
#if grt_vector_1 || grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b0).s0 |= (uint)charset[(z.s0 % 256) + charset_offset] << 24;
#endif
#if grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b0).s1 |= (uint)charset[(z.s1 % 256) + charset_offset] << 24;
#endif
#if grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b0).s2 |= (uint)charset[(z.s2 % 256) + charset_offset] << 24;
    (*b0).s3 |= (uint)charset[(z.s3 % 256) + charset_offset] << 24;
#endif
#if grt_vector_8 || grt_vector_16
    (*b0).s4 |= (uint)charset[(z.s4 % 256) + charset_offset] << 24;
    (*b0).s5 |= (uint)charset[(z.s5 % 256) + charset_offset] << 24;
    (*b0).s6 |= (uint)charset[(z.s6 % 256) + charset_offset] << 24;
    (*b0).s7 |= (uint)charset[(z.s7 % 256) + charset_offset] << 24;
#endif
    if (PasswordLength == 4) {return;}

    z /= (vector_type)256;

#if grt_vector_1 || grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b1).s0 = (uint)charset[(z.s0 % 256) + charset_offset];
#endif
#if grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b1).s1 = (uint)charset[(z.s1 % 256) + charset_offset];
#endif
#if grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b1).s2 = (uint)charset[(z.s2 % 256) + charset_offset];
    (*b1).s3 = (uint)charset[(z.s3 % 256) + charset_offset];
#endif
#if grt_vector_8 || grt_vector_16
    (*b1).s4 = (uint)charset[(z.s4 % 256) + charset_offset];
    (*b1).s5 = (uint)charset[(z.s5 % 256) + charset_offset];
    (*b1).s6 = (uint)charset[(z.s6 % 256) + charset_offset];
    (*b1).s7 = (uint)charset[(z.s7 % 256) + charset_offset];
#endif
    if (PasswordLength == 5) {return;}

    z /= (vector_type)256;
#if grt_vector_1 || grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b1).s0 |= (uint)charset[(z.s0 % 256) + charset_offset] << 8;
#endif
#if grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b1).s1 |= (uint)charset[(z.s1 % 256) + charset_offset] << 8;
#endif
#if grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b1).s2 |= (uint)charset[(z.s2 % 256) + charset_offset] << 8;
    (*b1).s3 |= (uint)charset[(z.s3 % 256) + charset_offset] << 8;
#endif
#if grt_vector_8 || grt_vector_16
    (*b1).s4 |= (uint)charset[(z.s4 % 256) + charset_offset] << 8;
    (*b1).s5 |= (uint)charset[(z.s5 % 256) + charset_offset] << 8;
    (*b1).s6 |= (uint)charset[(z.s6 % 256) + charset_offset] << 8;
    (*b1).s7 |= (uint)charset[(z.s7 % 256) + charset_offset] << 8;
#endif
    if (PasswordLength == 6) {return;}

    // Last 2
    z = (c+(vector_type)CurrentStep+(vector_type)Device_Table_Index) % (vector_type)(256*256*256);
#if grt_vector_1 || grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b1).s0 |= (uint)charset[(z.s0 % 256) + charset_offset] << 16;
#endif
#if grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b1).s1 |= (uint)charset[(z.s1 % 256) + charset_offset] << 16;
#endif
#if grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b1).s2 |= (uint)charset[(z.s2 % 256) + charset_offset] << 16;
    (*b1).s3 |= (uint)charset[(z.s3 % 256) + charset_offset] << 16;
#endif
#if grt_vector_8 || grt_vector_16
    (*b1).s4 |= (uint)charset[(z.s4 % 256) + charset_offset] << 16;
    (*b1).s5 |= (uint)charset[(z.s5 % 256) + charset_offset] << 16;
    (*b1).s6 |= (uint)charset[(z.s6 % 256) + charset_offset] << 16;
    (*b1).s7 |= (uint)charset[(z.s7 % 256) + charset_offset] << 16;
#endif
    if (PasswordLength == 7) {return;}

    z /= (vector_type)256;
#if grt_vector_1 || grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b1).s0 |= (uint)charset[(z.s0 % 256) + charset_offset] << 24;
#endif
#if grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b1).s1 |= (uint)charset[(z.s1 % 256) + charset_offset] << 24;
#endif
#if grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b1).s2 |= (uint)charset[(z.s2 % 256) + charset_offset] << 24;
    (*b1).s3 |= (uint)charset[(z.s3 % 256) + charset_offset] << 24;
#endif
#if grt_vector_8 || grt_vector_16
    (*b1).s4 |= (uint)charset[(z.s4 % 256) + charset_offset] << 24;
    (*b1).s5 |= (uint)charset[(z.s5 % 256) + charset_offset] << 24;
    (*b1).s6 |= (uint)charset[(z.s6 % 256) + charset_offset] << 24;
    (*b1).s7 |= (uint)charset[(z.s7 % 256) + charset_offset] << 24;
#endif
    if (PasswordLength == 8) {return;}

    z /= (vector_type)256;
#if grt_vector_1 || grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b2).s0 = (uint)charset[(z.s0 % 256) + charset_offset];
#endif
#if grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b2).s1 = (uint)charset[(z.s1 % 256) + charset_offset];
#endif
#if grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b2).s2 = (uint)charset[(z.s2 % 256) + charset_offset];
    (*b2).s3 = (uint)charset[(z.s3 % 256) + charset_offset];
#endif
#if grt_vector_8 || grt_vector_16
    (*b2).s4 = (uint)charset[(z.s4 % 256) + charset_offset];
    (*b2).s5 = (uint)charset[(z.s5 % 256) + charset_offset];
    (*b2).s6 = (uint)charset[(z.s6 % 256) + charset_offset];
    (*b2).s7 = (uint)charset[(z.s7 % 256) + charset_offset];
#endif
    if (PasswordLength == 9) {return;}

    z = (d+(vector_type)CurrentStep+(vector_type)Device_Table_Index) % (vector_type)(256*256*256);
#if grt_vector_1 || grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b2).s0 |= (uint)charset[(z.s0 % 256) + charset_offset] << 8;
#endif
#if grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b2).s1 |= (uint)charset[(z.s1 % 256) + charset_offset] << 8;
#endif
#if grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b2).s2 |= (uint)charset[(z.s2 % 256) + charset_offset] << 8;
    (*b2).s3 |= (uint)charset[(z.s3 % 256) + charset_offset] << 8;
#endif
#if grt_vector_8 || grt_vector_16
    (*b2).s4 |= (uint)charset[(z.s4 % 256) + charset_offset] << 8;
    (*b2).s5 |= (uint)charset[(z.s5 % 256) + charset_offset] << 8;
    (*b2).s6 |= (uint)charset[(z.s6 % 256) + charset_offset] << 8;
    (*b2).s7 |= (uint)charset[(z.s7 % 256) + charset_offset] << 8;
#endif
    if (PasswordLength == 10) {return;}

    z /= (vector_type)256;
#if grt_vector_1 || grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b2).s0 |= (uint)charset[(z.s0 % 256) + charset_offset] << 16;
#endif
#if grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b2).s1 |= (uint)charset[(z.s1 % 256) + charset_offset] << 16;
#endif
#if grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b2).s2 |= (uint)charset[(z.s2 % 256) + charset_offset] << 16;
    (*b2).s3 |= (uint)charset[(z.s3 % 256) + charset_offset] << 16;
#endif
#if grt_vector_8 || grt_vector_16
    (*b2).s4 |= (uint)charset[(z.s4 % 256) + charset_offset] << 16;
    (*b2).s5 |= (uint)charset[(z.s5 % 256) + charset_offset] << 16;
    (*b2).s6 |= (uint)charset[(z.s6 % 256) + charset_offset] << 16;
    (*b2).s7 |= (uint)charset[(z.s7 % 256) + charset_offset] << 16;
#endif
    if (PasswordLength == 11) {return;}

    z /= (vector_type)256;
#if grt_vector_1 || grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b2).s0 |= (uint)charset[(z.s0 % 256) + charset_offset] << 24;
#endif
#if grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b2).s1 |= (uint)charset[(z.s1 % 256) + charset_offset] << 24;
#endif
#if grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b2).s2 |= (uint)charset[(z.s2 % 256) + charset_offset] << 24;
    (*b2).s3 |= (uint)charset[(z.s3 % 256) + charset_offset] << 24;
#endif
#if grt_vector_8 || grt_vector_16
    (*b2).s4 |= (uint)charset[(z.s4 % 256) + charset_offset] << 24;
    (*b2).s5 |= (uint)charset[(z.s5 % 256) + charset_offset] << 24;
    (*b2).s6 |= (uint)charset[(z.s6 % 256) + charset_offset] << 24;
    (*b2).s7 |= (uint)charset[(z.s7 % 256) + charset_offset] << 24;
#endif
    if (PasswordLength == 12) {return;}
}


__kernel __attribute__((vec_type_hint(vector_type))) void CandidateHashMD5AMD(
    __constant unsigned char *deviceCharset,/* 0 */
    __private unsigned int deviceCharsetLength,/* 1 */
    __private unsigned int deviceChainLength,/* 2 */
    __private unsigned int deviceTableIndex,/* 3 */
    __private unsigned int deviceNumberOfThreads,/* 4 */
    __constant unsigned int *MD5_Candidate_Device_Hash,/* 5 */
    __global   unsigned int *outputCandidateHashArray, /* 6 */
    __private   unsigned int threadSpaceOffset,/* 7 */
    __private   unsigned int startStep,/* 8 */
    __private   unsigned int stepsToRun/* 9 */
) {

    // Start the kernel.
    __local unsigned char charset[512];

#if CPU_DEBUG
    printf("\n\n\n");
    printf("Kernel start, global id %d\n", get_global_id(0));
    printf("deviceCharsetLength: %d\n", deviceCharsetLength);
    printf("deviceCharset: %c %c %c %c ...\n", deviceCharset[0], deviceCharset[1], deviceCharset[2], deviceCharset[3]);
    printf("deviceChainLength: %d\n", deviceChainLength);
    printf("deviceTableIndex: %d\n", deviceTableIndex);
    printf("deviceNumberOfThreads: %d\n", deviceNumberOfThreads);
    printf("threadSpaceOffset: %d\n", threadSpaceOffset);
    printf("startStep: %d\n", startStep);
    printf("stepsToRun: %d\n", stepsToRun);
    printf("Candidate hash: %08x %08x %08x %08x\n", MD5_Candidate_Device_Hash[0],
            MD5_Candidate_Device_Hash[1], MD5_Candidate_Device_Hash[2], MD5_Candidate_Device_Hash[3]);
    printf("PASSWORD_LENGTH: %d\n", PASSWORD_LENGTH);
    for(int i = 0; i < 512; i++) {
        printf("%c.", deviceCharset[i]);
    }
#endif


    // Hash variables
    vector_type b0,b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15;
    vector_type a,b,c,d;
    // Vector to hold the initial values until we are done with them.
    // This is to load up the vector space nicely.
    uint init_a, init_b, init_c, init_d;

    uint i, chain_index, step_to_calculate, charset_offset, last_step_for_iteration;

    // Generic "copy charset to shared memory" function
    copySingleCharsetToShared(charset, deviceCharset);
    //printf("Charset copied... %c %c %c %c ...\n", charset[0], charset[1], charset[2], charset[3]);

    chain_index = (get_global_id(0) * VECTOR_WIDTH + (threadSpaceOffset * deviceNumberOfThreads));

#if CPU_DEBUG
    printf("chain index: %d\n", chain_index);
#endif

    // Find out if we're done with work.
    // If our index + the startstep is greater than the chain length, this thread has nothing to do.
    if ((chain_index) >= deviceChainLength) {
#if CPU_DEBUG
        printf("returning: chain_index %d  start_step %d  chain length %d \n", chain_index, startStep, deviceChainLength);
#endif
        return;
    }



    // Figure out which step we're running.
    step_to_calculate = chain_index + startStep;
#if CPU_DEBUG
    printf("Step to calculate: %d \n", step_to_calculate);
#endif

    // Yes, modulus here is slow.  And this is a less-critical chunk of code.  So it stays here.
    charset_offset = step_to_calculate % deviceCharsetLength;

    b0 = 0x00000000;
    b1 = 0x00000000;
    b2 = 0x00000000;
    b3 = 0x00000000;
    b4 = 0x00000000;
    b5 = 0x00000000;
    b6 = 0x00000000;
    b7 = 0x00000000;
    b8 = 0x00000000;
    b9 = 0x00000000;
    b10 = 0x00000000;
    b11 = 0x00000000;
    b12 = 0x00000000;
    b13 = 0x00000000;
    b14 = 0x00000000;
    b15 = 0x00000000;

    // Load the initial hash.  This will either be from the constant or from the storage space.
    // If it is from the constant area, load it into the init_{a,b,c,d} vars so we can copy
    // it as needed.
    if (startStep == 0) {
        init_a = (uint)MD5_Candidate_Device_Hash[0];
        init_b = (uint)MD5_Candidate_Device_Hash[1];
        init_c = (uint)MD5_Candidate_Device_Hash[2];
        init_d = (uint)MD5_Candidate_Device_Hash[3];
#if CPU_DEBUG
        printf("Loaded candidate %08x %08x %08x %08x \n", init_a, init_b, init_c, init_d);
#endif
        a = (vector_type) 0;
        b = (vector_type) 0;
        c = (vector_type) 0;
        d = (vector_type) 0;

    // Now step through the vectors to load them properly...
 #if grt_vector_1 || grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
        a.s0 = init_a;
        b.s0 = init_b;
        c.s0 = init_c;
        d.s0 = init_d;
        reduceSingleCharsetNormal(&b0, &b1, &b2, a, b, c, d, step_to_calculate, charset, charset_offset, PASSWORD_LENGTH, deviceTableIndex);
    #if CPU_DEBUG
        printf("Reduced password .s0: %08x %08x %08x\n", b0.s0, b1.s0, b2.s0);
    #endif
        charset_offset++;
        if (charset_offset >= deviceCharsetLength) {
            charset_offset = 0;
        }
        step_to_calculate++;
        stepsToRun--;
        padMDHash(PASSWORD_LENGTH, &b0, &b1, &b2, &b3, &b4, &b5, &b6, &b7, &b8, &b9, &b10, &b11, &b12, &b13, &b14, &b15);
        OpenCL_MD5(b0, b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15, &a, &b, &c, &d);
        a.s1 = init_a;
        b.s1 = init_b;
        c.s1 = init_c;
        d.s1 = init_d;
    #endif
#if grt_vector_4 || grt_vector_8 || grt_vector_16
        reduceSingleCharsetNormal(&b0, &b1, &b2, a, b, c, d, step_to_calculate, charset, charset_offset, PASSWORD_LENGTH, deviceTableIndex);
    #if CPU_DEBUG
        printf("Reduced password .s0: %08x %08x %08x\n", b0.s0, b1.s0, b2.s0);
    #endif
        charset_offset++;
        if (charset_offset >= deviceCharsetLength) {
            charset_offset = 0;
        }
        step_to_calculate++;
        stepsToRun--;
        padMDHash(PASSWORD_LENGTH, &b0, &b1, &b2, &b3, &b4, &b5, &b6, &b7, &b8, &b9, &b10, &b11, &b12, &b13, &b14, &b15);
        OpenCL_MD5(b0, b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15, &a, &b, &c, &d);
        a.s2 = init_a;
        b.s2 = init_b;
        c.s2 = init_c;
        d.s2 = init_d;
        reduceSingleCharsetNormal(&b0, &b1, &b2, a, b, c, d, step_to_calculate, charset, charset_offset, PASSWORD_LENGTH, deviceTableIndex);
    #if CPU_DEBUG
        printf("Reduced password .s0: %08x %08x %08x\n", b0.s0, b1.s0, b2.s0);
    #endif
        charset_offset++;
        if (charset_offset >= deviceCharsetLength) {
            charset_offset = 0;
        }
        step_to_calculate++;
        stepsToRun--;
        padMDHash(PASSWORD_LENGTH, &b0, &b1, &b2, &b3, &b4, &b5, &b6, &b7, &b8, &b9, &b10, &b11, &b12, &b13, &b14, &b15);
        OpenCL_MD5(b0, b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15, &a, &b, &c, &d);
        a.s3 = init_a;
        b.s3 = init_b;
        c.s3 = init_c;
        d.s3 = init_d;
    #endif

    } else {
        a = vload_type( 0 * deviceChainLength                 + chain_index / VECTOR_WIDTH, outputCandidateHashArray);
        b = vload_type((1 * deviceChainLength / VECTOR_WIDTH) + chain_index / VECTOR_WIDTH, outputCandidateHashArray);
        c = vload_type((2 * deviceChainLength / VECTOR_WIDTH) + chain_index / VECTOR_WIDTH, outputCandidateHashArray);
        d = vload_type((3 * deviceChainLength / VECTOR_WIDTH) + chain_index / VECTOR_WIDTH, outputCandidateHashArray);
#if CPU_DEBUG
    printf("Loaded hash from location %d .s0:  %08x %08x %08x %08x \n", chain_index,
            a.s0, b.s0, c.s0, d.s0);
    printf("Loaded hash from location %d .s1:  %08x %08x %08x %08x \n", chain_index,
            a.s1, b.s1, c.s1, d.s1);
#endif

    }




#if CPU_DEBUG
    printf("Entering reduceCharset...\n");
    printf("Charset offset: %d\n", charset_offset);
#endif

    reduceSingleCharsetNormal(&b0, &b1, &b2, a, b, c, d, step_to_calculate, charset, charset_offset, PASSWORD_LENGTH, deviceTableIndex);
#if CPU_DEBUG
    printf("Reduced password .s0: %08x %08x %08x\n", b0.s0, b1.s0, b2.s0);
    printf("Reduced password .s1: %08x %08x %08x\n", b0.s1, b1.s1, b2.s1);
#endif

    step_to_calculate++;
    charset_offset++;
    if (charset_offset >= deviceCharsetLength) {
        charset_offset = 0;
    }
    // Figure out the last step to run - either the chain length or
    // the number of specified steps.
    if ((step_to_calculate + stepsToRun) > deviceChainLength) {
        last_step_for_iteration = deviceChainLength - 1;
    } else {
        last_step_for_iteration = (step_to_calculate + stepsToRun - 1); // Already run one
    }
#if CPU_DEBUG
    printf("last_step_for_iteration: %d \n", last_step_for_iteration);
#endif


    // We now have our (step+1) charset.
    for (i = step_to_calculate; i <= last_step_for_iteration; i++) {
#if CPU_DEBUG
    printf("Chain step %d \n", i);
#endif
        padMDHash(PASSWORD_LENGTH, &b0, &b1, &b2, &b3, &b4, &b5, &b6, &b7, &b8, &b9, &b10, &b11, &b12, &b13, &b14, &b15);
        OpenCL_MD5(b0, b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15, &a, &b, &c, &d);
#if CPU_DEBUG
    printf("MD5 result .s0: %08x %08x %08x %08x \n", a.s0, b.s0, c.s0, d.s0);
    printf("MD5 result .s1: %08x %08x %08x %08x \n", a.s1, b.s1, c.s1, d.s1);
#endif
        reduceSingleCharsetNormal(&b0, &b1, &b2, a, b, c, d, i, charset, charset_offset, PASSWORD_LENGTH, deviceTableIndex);
#if CPU_DEBUG
    printf("Reduced password .s0: %08x %08x %08x\n", b0.s0, b1.s0, b2.s0);
    printf("Reduced password .s1: %08x %08x %08x\n", b0.s1, b1.s1, b2.s1);
#endif
        charset_offset++;
        if (charset_offset >= deviceCharsetLength) {
            charset_offset = 0;
        }
    }
    // Store the hash output.
#if CPU_DEBUG
    if (chain_index > deviceChainLength) {
        printf("ERROR: SHOULD NOT WRITE THIS!!!  Chain index to store: %d\n", chain_index);
    }
#endif
    vstore_type(a,  0 * deviceChainLength                 + chain_index / VECTOR_WIDTH, outputCandidateHashArray);
    vstore_type(b, (1 * deviceChainLength / VECTOR_WIDTH) + chain_index / VECTOR_WIDTH, outputCandidateHashArray);
    vstore_type(c, (2 * deviceChainLength / VECTOR_WIDTH) + chain_index / VECTOR_WIDTH, outputCandidateHashArray);
    vstore_type(d, (3 * deviceChainLength / VECTOR_WIDTH) + chain_index / VECTOR_WIDTH, outputCandidateHashArray);
#if CPU_DEBUG
    printf("Stored candidate .s0 %08x %08x %08x %08x \n", a.s0, b.s0, c.s0, d.s0);
    printf("Stored candidate .s1 %08x %08x %08x %08x \n", a.s1, b.s1, c.s1, d.s1);
#endif
}


