// Kernels for MD5 - vectorized for AMD...

// Things we should define in the calling code...
//#define VECTOR_WIDTH 4
//#define CPU_DEBUG 1
//#define BITALIGN
//#define NVIDIA_HACKS 1
//#define PASSWORD_LENGTH 6
//#define INTEL_OPENCL_HACK 1


#if VECTOR_WIDTH == 1
    #define vector_type uint
    #define vload_type vload1
    #define vstore_type vstore1
    #define grt_vector_1 1
#elif VECTOR_WIDTH == 2
    #define vector_type uint2
    #define vload_type vload2
    #define vstore_type vstore2
    #define grt_vector_2 1
#elif VECTOR_WIDTH == 4
    #define vector_type uint4
    #define vload_type vload4
    #define vstore_type vstore4
    #define grt_vector_4 1
#elif VECTOR_WIDTH == 8
ERROR NOT SUPPORTED
    #define vector_type uint8
    #define vload_type vload8
    #define vstore_type vstore8
    #define grt_vector_8 1
#elif VECTOR_WIDTH == 16
ERROR NOT SUPPORTED
    #define vector_type uint16
    #define vload_type vload16
    #define vstore_type vstore16
    #define grt_vector_16 1
#endif


#ifdef INTEL_OPENCL_HACK
#define LOCAL_MEMORY_TYPE __constant
#define CHARSET_NAME deviceCharset
#define BITMAP_NAME constantBitmap
#else
#define LOCAL_MEMORY_TYPE __local
#define CHARSET_NAME charset
#define BITMAP_NAME sharedBitmap
#endif

#ifdef CPU_DEBUG
#pragma OPENCL EXTENSION cl_amd_printf : enable
#endif

#define MAX_PASSWORD_LENGTH 32

// Hash defines

#define MD5S11 7
#define MD5S12 12
#define MD5S13 17
#define MD5S14 22
#define MD5S21 5
#define MD5S22 9
#define MD5S23 14
#define MD5S24 20
#define MD5S31 4
#define MD5S32 11
#define MD5S33 16
#define MD5S34 23
#define MD5S41 6
#define MD5S42 10
#define MD5S43 15
#define MD5S44 21

// OPTIMIZED MD5 FUNCTIONS HERE
//#define MD5F(x,y,z) (((y ^ z) & x) ^ z)
//#define MD5G(x,y,z) (((x & y) & z) ^ y)

/* F, G, H and I are basic MD5 functions.
 */
#define MD5F(x, y, z) (((x) & (y)) | ((~x) & (z)))
#define MD5G(x, y, z) (((x) & (z)) | ((y) & (~z)))
#define MD5H(x, y, z) ((x) ^ (y) ^ (z))
#define MD5I(x, y, z) ((y) ^ ((x) | (~z)))


/* ROTATE_LEFT rotates x left n bits.
 */

#ifdef BITALIGN
#pragma OPENCL EXTENSION cl_amd_media_ops : enable
#define MD5ROTATE_LEFT(x, y) amd_bitalign(x, x, (uint)(32 - y))
#define MD5FF(a, b, c, d, x, s, ac) { \
 (a) += amd_bytealign((b),(c),(d)) + (x) + (vector_type)(ac); \
 (a) = MD5ROTATE_LEFT ((a), (s)); \
 (a) += (b); \
  }

#elif NVIDIA_HACKS
#define MD5ROTATE_LEFT(x, n) (((x) << (n)) | ((x) >> (32-(n))))
#define MD5FF(a, b, c, d, x, s, ac) { \
 (a) += MD5F ((b), (c), (d)) + (x) + (vector_type)(ac); \
 (a) = MD5ROTATE_LEFT ((a), (s)); \
 (a) += (b); \
  }
#else
#define MD5ROTATE_LEFT(x, y) rotate((vector_type)x, (uint)y)
#define MD5FF(a, b, c, d, x, s, ac) { \
 (a) += MD5F ((b), (c), (d)) + (x) + (vector_type)(ac); \
 (a) = MD5ROTATE_LEFT ((a), (s)); \
 (a) += (b); \
  }
#endif


//#define MD5ROTATE_LEFT(x, n) (((x) << (n)) | ((x) >> (32-(n))))

/* FF, GG, HH, and II transformations for rounds 1, 2, 3, and 4.
Rotation is separate from addition to prevent recomputation.
 */

/*#define MD5FF(a, b, c, d, x, s, ac) { \
 (a) += MD5F ((b), (c), (d)) + (x) + (vector_type)(ac); \
 (a) = MD5ROTATE_LEFT ((a), (s)); \
 (a) += (b); \
  }*/
/*
#pragma OPENCL EXTENSION cl_amd_media_ops : enable

#define MD5FF(a, b, c, d, x, s, ac) { \
 (a) += amd_bytealign((b),(c),(d)) + (x) + (vector_type)(ac); \
 (a) = MD5ROTATE_LEFT ((a), (s)); \
 (a) += (b); \
  }
*/
#define MD5GG(a, b, c, d, x, s, ac) { \
 (a) += MD5G ((b), (c), (d)) + (x) + (vector_type)(ac); \
 (a) = MD5ROTATE_LEFT ((a), (s)); \
 (a) += (b); \
  }
#define MD5HH(a, b, c, d, x, s, ac) { \
 (a) += MD5H ((b), (c), (d)) + (x) + (vector_type)(ac); \
 (a) = MD5ROTATE_LEFT ((a), (s)); \
 (a) += (b); \
  }
#define MD5II(a, b, c, d, x, s, ac) { \
 (a) += MD5I ((b), (c), (d)) + (x) + (vector_type)(ac); \
 (a) = MD5ROTATE_LEFT ((a), (s)); \
 (a) += (b); \
  }


inline void OpenCL_MD5(vector_type b0, vector_type b1, vector_type b2, vector_type b3, vector_type b4, vector_type b5, vector_type b6, vector_type b7,
			   vector_type b8, vector_type b9, vector_type b10, vector_type b11, vector_type b12, vector_type b13, vector_type b14, vector_type b15,
			   vector_type *a, vector_type *b, vector_type *c, vector_type *d) {
  // Load these with vectors.
  *a = (vector_type)0x67452301;
  *b = (vector_type)0xefcdab89;
  *c = (vector_type)0x98badcfe;
  *d = (vector_type)0x10325476;

  MD5FF (*a, *b, *c, *d, b0, MD5S11, 0xd76aa478); /* 1 */
  MD5FF (*d, *a, *b, *c, b1, MD5S12, 0xe8c7b756); /* 2 */
  MD5FF (*c, *d, *a, *b, b2, MD5S13, 0x242070db); /* 3 */
  MD5FF (*b, *c, *d, *a, b3, MD5S14, 0xc1bdceee); /* 4 */
  MD5FF (*a, *b, *c, *d, b4, MD5S11, 0xf57c0faf); /* 5 */
  MD5FF (*d, *a, *b, *c, b5, MD5S12, 0x4787c62a); /* 6 */
  MD5FF (*c, *d, *a, *b, b6, MD5S13, 0xa8304613); /* 7 */
  MD5FF (*b, *c, *d, *a, b7, MD5S14, 0xfd469501); /* 8 */
  MD5FF (*a, *b, *c, *d, b8, MD5S11, 0x698098d8); /* 9 */
  MD5FF (*d, *a, *b, *c, b9, MD5S12, 0x8b44f7af); /* 10 */
  MD5FF (*c, *d, *a, *b, b10, MD5S13, 0xffff5bb1); /* 11 */
  MD5FF (*b, *c, *d, *a, b11, MD5S14, 0x895cd7be); /* 12 */
  MD5FF (*a, *b, *c, *d, b12, MD5S11, 0x6b901122); /* 13 */
  MD5FF (*d, *a, *b, *c, b13, MD5S12, 0xfd987193); /* 14 */
  MD5FF (*c, *d, *a, *b, b14, MD5S13, 0xa679438e); /* 15 */
  MD5FF (*b, *c, *d, *a, b15, MD5S14, 0x49b40821); /* 16 */

 /* Round 2 */
  MD5GG (*a, *b, *c, *d, b1, MD5S21, 0xf61e2562); /* 17 */
  MD5GG (*d, *a, *b, *c, b6, MD5S22, 0xc040b340); /* 18 */
  MD5GG (*c, *d, *a, *b, b11, MD5S23, 0x265e5a51); /* 19 */
  MD5GG (*b, *c, *d, *a, b0, MD5S24, 0xe9b6c7aa); /* 20 */
  MD5GG (*a, *b, *c, *d, b5, MD5S21, 0xd62f105d); /* 21 */
  MD5GG (*d, *a, *b, *c, b10, MD5S22,  0x2441453); /* 22 */
  MD5GG (*c, *d, *a, *b, b15, MD5S23, 0xd8a1e681); /* 23 */
  MD5GG (*b, *c, *d, *a, b4, MD5S24, 0xe7d3fbc8); /* 24 */
  MD5GG (*a, *b, *c, *d, b9, MD5S21, 0x21e1cde6); /* 25 */
  MD5GG (*d, *a, *b, *c, b14, MD5S22, 0xc33707d6); /* 26 */
  MD5GG (*c, *d, *a, *b, b3, MD5S23, 0xf4d50d87); /* 27 */
  MD5GG (*b, *c, *d, *a, b8, MD5S24, 0x455a14ed); /* 28 */
  MD5GG (*a, *b, *c, *d, b13, MD5S21, 0xa9e3e905); /* 29 */
  MD5GG (*d, *a, *b, *c, b2, MD5S22, 0xfcefa3f8); /* 30 */
  MD5GG (*c, *d, *a, *b, b7, MD5S23, 0x676f02d9); /* 31 */
  MD5GG (*b, *c, *d, *a, b12, MD5S24, 0x8d2a4c8a); /* 32 */

  /* Round 3 */
  MD5HH (*a, *b, *c, *d, b5, MD5S31, 0xfffa3942); /* 33 */
  MD5HH (*d, *a, *b, *c, b8, MD5S32, 0x8771f681); /* 34 */
  MD5HH (*c, *d, *a, *b, b11, MD5S33, 0x6d9d6122); /* 35 */
  MD5HH (*b, *c, *d, *a, b14, MD5S34, 0xfde5380c); /* 36 */
  MD5HH (*a, *b, *c, *d, b1, MD5S31, 0xa4beea44); /* 37 */
  MD5HH (*d, *a, *b, *c, b4, MD5S32, 0x4bdecfa9); /* 38 */
  MD5HH (*c, *d, *a, *b, b7, MD5S33, 0xf6bb4b60); /* 39 */
  MD5HH (*b, *c, *d, *a, b10, MD5S34, 0xbebfbc70); /* 40 */
  MD5HH (*a, *b, *c, *d, b13, MD5S31, 0x289b7ec6); /* 41 */
  MD5HH (*d, *a, *b, *c, b0, MD5S32, 0xeaa127fa); /* 42 */
  MD5HH (*c, *d, *a, *b, b3, MD5S33, 0xd4ef3085); /* 43 */
  MD5HH (*b, *c, *d, *a, b6, MD5S34,  0x4881d05); /* 44 */
  MD5HH (*a, *b, *c, *d, b9, MD5S31, 0xd9d4d039); /* 45 */
  MD5HH (*d, *a, *b, *c, b12, MD5S32, 0xe6db99e5); /* 46 */
  MD5HH (*c, *d, *a, *b, b15, MD5S33, 0x1fa27cf8); /* 47 */
  MD5HH (*b, *c, *d, *a, b2, MD5S34, 0xc4ac5665); /* 48 */

  /* Round 4 */
  MD5II (*a, *b, *c, *d, b0, MD5S41, 0xf4292244); /* 49 */
  MD5II (*d, *a, *b, *c, b7, MD5S42, 0x432aff97); /* 50 */
  MD5II (*c, *d, *a, *b, b14, MD5S43, 0xab9423a7); /* 51 */
  MD5II (*b, *c, *d, *a, b5, MD5S44, 0xfc93a039); /* 52 */
  MD5II (*a, *b, *c, *d, b12, MD5S41, 0x655b59c3); /* 53 */
  MD5II (*d, *a, *b, *c, b3, MD5S42, 0x8f0ccc92); /* 54 */
  MD5II (*c, *d, *a, *b, b10, MD5S43, 0xffeff47d); /* 55 */
  MD5II (*b, *c, *d, *a, b1, MD5S44, 0x85845dd1); /* 56 */
  MD5II (*a, *b, *c, *d, b8, MD5S41, 0x6fa87e4f); /* 57 */
  MD5II (*d, *a, *b, *c, b15, MD5S42, 0xfe2ce6e0); /* 58 */
  MD5II (*c, *d, *a, *b, b6, MD5S43, 0xa3014314); /* 59 */
  MD5II (*b, *c, *d, *a, b13, MD5S44, 0x4e0811a1); /* 60 */
  MD5II (*a, *b, *c, *d, b4, MD5S41, 0xf7537e82); /* 61 */
  MD5II (*d, *a, *b, *c, b11, MD5S42, 0xbd3af235); /* 62 */
  MD5II (*c, *d, *a, *b, b2, MD5S43, 0x2ad7d2bb); /* 63 */
  MD5II (*b, *c, *d, *a, b9, MD5S44, 0xeb86d391); /* 64 */

  // Finally, add initial values, as this is the only pass we make.
  *a += (vector_type)0x67452301;
  *b += (vector_type)0xefcdab89;
  *c += (vector_type)0x98badcfe;
  *d += (vector_type)0x10325476;
}


inline void copySingleCharsetToShared(__local unsigned char *sharedCharset,
        __constant unsigned char *constantCharset) {

    // 32-bit accesses are used to help coalesce memory accesses.
    uint a;

    for (a = 0; a < (512); a++) {
        sharedCharset[a] = constantCharset[a];
    }
}

inline void copySingleBitmapToShared(__local unsigned char *sharedBitmap,
        __constant unsigned char *globalBitmap) {

    // 32-bit accesses are used to help coalesce memory accesses.
    uint a;

    for (a = 0; a < (8192); a++) {
        sharedBitmap[a] = globalBitmap[a];
    }
}


inline void padMDHash(int length,
                vector_type *b0, vector_type *b1, vector_type *b2, vector_type *b3, vector_type *b4, vector_type *b5, vector_type *b6, vector_type *b7,
		vector_type *b8, vector_type *b9, vector_type *b10, vector_type *b11, vector_type *b12, vector_type *b13, vector_type *b14, vector_type *b15) {

  // Set length properly (length in bits)
  *b14 = (vector_type)(length * 8);

  if (length == 0) {
    *b0 |= (vector_type)0x00000080;
    return;
  }
  if (length == 1) {
    *b0 |= (vector_type)0x00008000;
    return;
  }
  if (length == 2) {
    *b0 |= (vector_type)0x00800000;
    return;
  }
  if (length == 3) {
    *b0 |= (vector_type)0x80000000;
    return;
  }
  if (length == 4) {
    *b1 |= (vector_type)0x00000080;
    return;
  }
  if (length == 5) {
    *b1 |= (vector_type)0x00008000;
    return;
  }
  if (length == 6) {
    *b1 |= (vector_type)0x00800000;
    return;
  }
  if (length == 7) {
    *b1 |= (vector_type)0x80000000;
    return;
  }
  if (length == 8) {
    *b2 |= (vector_type)0x00000080;
    return;
  }
  if (length == 9) {
    *b2 |= (vector_type)0x00008000;
    return;
  }
  if (length == 10) {
    *b2 |= (vector_type)0x00800000;
    return;
  }
  if (length == 11) {
   *b2 |= (vector_type)0x80000000;
    return;
  }
  if (length == 12) {
    *b3 |= (vector_type)0x00000080;
    return;
  }
  if (length == 13) {
    *b3 |= (vector_type)0x00008000;
    return;
  }
  if (length == 14) {
    *b3 |= (vector_type)0x00800000;
    return;
  }
  if (length == 15) {
    *b3 |= (vector_type)0x80000000;
    return;
  }
  if (length == 16) {
    *b4 |= (vector_type)0x00000080;
    return;
  }
  if (length == 17) {
    *b4 |= (vector_type)0x00008000;
    return;
  }
  if (length == 18) {
    *b4 |= (vector_type)0x00800000;
    return;
  }
  if (length == 19) {
    *b4 |= (vector_type)0x80000000;
    return;
  }
  if (length == 20) {
    *b5 |= (vector_type)0x00000080;
    return;
  }
  if (length == 21) {
    *b5 |= (vector_type)0x00008000;
    return;
  }
  if (length == 22) {
    *b5 |= (vector_type)0x00800000;
    return;
  }
  if (length == 23) {
    *b5 |= (vector_type)0x80000000;
    return;
  }
  if (length == 24) {
    *b6 |= (vector_type)0x00000080;
    return;
  }
  if (length == 25) {
    *b6 |= (vector_type)0x00008000;
    return;
  }
  if (length == 26) {
    *b6 |= (vector_type)0x00800000;
    return;
  }
  if (length == 27) {
    *b6 |= (vector_type)0x80000000;
    return;
  }
  if (length == 28) {
    *b7 |= (vector_type)0x00000080;
    return;
  }
  if (length == 29) {
    *b7 |= (vector_type)0x00008000;
    return;
  }
  if (length == 30) {
    *b7 |= (vector_type)0x00800000;
    return;
  }
  if (length == 31) {
    *b7 |= (vector_type)0x80000000;
    return;
  }

}

inline void reduceSingleCharsetNormal(vector_type *b0, vector_type *b1, vector_type *b2,
        vector_type a, vector_type b, vector_type c, vector_type d,
        uint CurrentStep, LOCAL_MEMORY_TYPE unsigned char *charset, uint charset_offset, int PasswordLength, uint Device_Table_Index) {

    vector_type z;
    // Reduce it
    // First 3
#if CPU_DEBUG && 0
    for (int i = 0; i < 512; i++) {
        printf("Charset %d: 0x%08x %c\n", i, charset[i], charset[i]);
    }
#endif
    z = (a+(vector_type)CurrentStep+(vector_type)Device_Table_Index) % (vector_type)(256*256*256);

#if grt_vector_1 || grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b0).s0 = (uint)charset[(z.s0 % 256) + charset_offset];
#endif
#if grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b0).s1 = (uint)charset[(z.s1 % 256) + charset_offset];
#endif
#if grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b0).s2 = (uint)charset[(z.s2 % 256) + charset_offset];
    (*b0).s3 = (uint)charset[(z.s3 % 256) + charset_offset];
#endif
#if grt_vector_8 || grt_vector_16
    (*b0).s4 = (uint)charset[(z.s4 % 256) + charset_offset];
    (*b0).s5 = (uint)charset[(z.s5 % 256) + charset_offset];
    (*b0).s6 = (uint)charset[(z.s6 % 256) + charset_offset];
    (*b0).s7 = (uint)charset[(z.s7 % 256) + charset_offset];
#endif
    if (PasswordLength == 1) {return;}

    z /= (vector_type)256;
#if grt_vector_1 || grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b0).s0 |= (uint)charset[(z.s0 % 256) + charset_offset] << 8;
#endif
#if grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b0).s1 |= (uint)charset[(z.s1 % 256) + charset_offset] << 8;
#endif
#if grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b0).s2 |= (uint)charset[(z.s2 % 256) + charset_offset] << 8;
    (*b0).s3 |= (uint)charset[(z.s3 % 256) + charset_offset] << 8;
#endif
#if grt_vector_8 || grt_vector_16
    (*b0).s4 |= (uint)charset[(z.s4 % 256) + charset_offset] << 8;
    (*b0).s5 |= (uint)charset[(z.s5 % 256) + charset_offset] << 8;
    (*b0).s6 |= (uint)charset[(z.s6 % 256) + charset_offset] << 8;
    (*b0).s7 |= (uint)charset[(z.s7 % 256) + charset_offset] << 8;
#endif
    if (PasswordLength == 2) {return;}

    z /= (vector_type)256;
#if grt_vector_1 || grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b0).s0 |= (uint)charset[(z.s0 % 256) + charset_offset] << 16;
#endif
#if grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b0).s1 |= (uint)charset[(z.s1 % 256) + charset_offset] << 16;
#endif
#if grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b0).s2 |= (uint)charset[(z.s2 % 256) + charset_offset] << 16;
    (*b0).s3 |= (uint)charset[(z.s3 % 256) + charset_offset] << 16;
#endif
#if grt_vector_8 || grt_vector_16
    (*b0).s4 |= (uint)charset[(z.s4 % 256) + charset_offset] << 16;
    (*b0).s5 |= (uint)charset[(z.s5 % 256) + charset_offset] << 16;
    (*b0).s6 |= (uint)charset[(z.s6 % 256) + charset_offset] << 16;
    (*b0).s7 |= (uint)charset[(z.s7 % 256) + charset_offset] << 16;
#endif
    if (PasswordLength == 3) {return;}

    // Second 3
    z = (b+(vector_type)CurrentStep+(vector_type)Device_Table_Index) % (vector_type)(256*256*256);
#if grt_vector_1 || grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b0).s0 |= (uint)charset[(z.s0 % 256) + charset_offset] << 24;
#endif
#if grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b0).s1 |= (uint)charset[(z.s1 % 256) + charset_offset] << 24;
#endif
#if grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b0).s2 |= (uint)charset[(z.s2 % 256) + charset_offset] << 24;
    (*b0).s3 |= (uint)charset[(z.s3 % 256) + charset_offset] << 24;
#endif
#if grt_vector_8 || grt_vector_16
    (*b0).s4 |= (uint)charset[(z.s4 % 256) + charset_offset] << 24;
    (*b0).s5 |= (uint)charset[(z.s5 % 256) + charset_offset] << 24;
    (*b0).s6 |= (uint)charset[(z.s6 % 256) + charset_offset] << 24;
    (*b0).s7 |= (uint)charset[(z.s7 % 256) + charset_offset] << 24;
#endif
    if (PasswordLength == 4) {return;}

    z /= (vector_type)256;

#if grt_vector_1 || grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b1).s0 = (uint)charset[(z.s0 % 256) + charset_offset];
#endif
#if grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b1).s1 = (uint)charset[(z.s1 % 256) + charset_offset];
#endif
#if grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b1).s2 = (uint)charset[(z.s2 % 256) + charset_offset];
    (*b1).s3 = (uint)charset[(z.s3 % 256) + charset_offset];
#endif
#if grt_vector_8 || grt_vector_16
    (*b1).s4 = (uint)charset[(z.s4 % 256) + charset_offset];
    (*b1).s5 = (uint)charset[(z.s5 % 256) + charset_offset];
    (*b1).s6 = (uint)charset[(z.s6 % 256) + charset_offset];
    (*b1).s7 = (uint)charset[(z.s7 % 256) + charset_offset];
#endif
    if (PasswordLength == 5) {return;}

    z /= (vector_type)256;
#if grt_vector_1 || grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b1).s0 |= (uint)charset[(z.s0 % 256) + charset_offset] << 8;
#endif
#if grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b1).s1 |= (uint)charset[(z.s1 % 256) + charset_offset] << 8;
#endif
#if grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b1).s2 |= (uint)charset[(z.s2 % 256) + charset_offset] << 8;
    (*b1).s3 |= (uint)charset[(z.s3 % 256) + charset_offset] << 8;
#endif
#if grt_vector_8 || grt_vector_16
    (*b1).s4 |= (uint)charset[(z.s4 % 256) + charset_offset] << 8;
    (*b1).s5 |= (uint)charset[(z.s5 % 256) + charset_offset] << 8;
    (*b1).s6 |= (uint)charset[(z.s6 % 256) + charset_offset] << 8;
    (*b1).s7 |= (uint)charset[(z.s7 % 256) + charset_offset] << 8;
#endif
    if (PasswordLength == 6) {return;}

    // Last 2
    z = (c+(vector_type)CurrentStep+(vector_type)Device_Table_Index) % (vector_type)(256*256*256);
#if grt_vector_1 || grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b1).s0 |= (uint)charset[(z.s0 % 256) + charset_offset] << 16;
#endif
#if grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b1).s1 |= (uint)charset[(z.s1 % 256) + charset_offset] << 16;
#endif
#if grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b1).s2 |= (uint)charset[(z.s2 % 256) + charset_offset] << 16;
    (*b1).s3 |= (uint)charset[(z.s3 % 256) + charset_offset] << 16;
#endif
#if grt_vector_8 || grt_vector_16
    (*b1).s4 |= (uint)charset[(z.s4 % 256) + charset_offset] << 16;
    (*b1).s5 |= (uint)charset[(z.s5 % 256) + charset_offset] << 16;
    (*b1).s6 |= (uint)charset[(z.s6 % 256) + charset_offset] << 16;
    (*b1).s7 |= (uint)charset[(z.s7 % 256) + charset_offset] << 16;
#endif
    if (PasswordLength == 7) {return;}

    z /= (vector_type)256;
#if grt_vector_1 || grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b1).s0 |= (uint)charset[(z.s0 % 256) + charset_offset] << 24;
#endif
#if grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b1).s1 |= (uint)charset[(z.s1 % 256) + charset_offset] << 24;
#endif
#if grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b1).s2 |= (uint)charset[(z.s2 % 256) + charset_offset] << 24;
    (*b1).s3 |= (uint)charset[(z.s3 % 256) + charset_offset] << 24;
#endif
#if grt_vector_8 || grt_vector_16
    (*b1).s4 |= (uint)charset[(z.s4 % 256) + charset_offset] << 24;
    (*b1).s5 |= (uint)charset[(z.s5 % 256) + charset_offset] << 24;
    (*b1).s6 |= (uint)charset[(z.s6 % 256) + charset_offset] << 24;
    (*b1).s7 |= (uint)charset[(z.s7 % 256) + charset_offset] << 24;
#endif
    if (PasswordLength == 8) {return;}

    z /= (vector_type)256;
#if grt_vector_1 || grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b2).s0 = (uint)charset[(z.s0 % 256) + charset_offset];
#endif
#if grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b2).s1 = (uint)charset[(z.s1 % 256) + charset_offset];
#endif
#if grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b2).s2 = (uint)charset[(z.s2 % 256) + charset_offset];
    (*b2).s3 = (uint)charset[(z.s3 % 256) + charset_offset];
#endif
#if grt_vector_8 || grt_vector_16
    (*b2).s4 = (uint)charset[(z.s4 % 256) + charset_offset];
    (*b2).s5 = (uint)charset[(z.s5 % 256) + charset_offset];
    (*b2).s6 = (uint)charset[(z.s6 % 256) + charset_offset];
    (*b2).s7 = (uint)charset[(z.s7 % 256) + charset_offset];
#endif
    if (PasswordLength == 9) {return;}

    z = (d+(vector_type)CurrentStep+(vector_type)Device_Table_Index) % (vector_type)(256*256*256);
#if grt_vector_1 || grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b2).s0 |= (uint)charset[(z.s0 % 256) + charset_offset] << 8;
#endif
#if grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b2).s1 |= (uint)charset[(z.s1 % 256) + charset_offset] << 8;
#endif
#if grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b2).s2 |= (uint)charset[(z.s2 % 256) + charset_offset] << 8;
    (*b2).s3 |= (uint)charset[(z.s3 % 256) + charset_offset] << 8;
#endif
#if grt_vector_8 || grt_vector_16
    (*b2).s4 |= (uint)charset[(z.s4 % 256) + charset_offset] << 8;
    (*b2).s5 |= (uint)charset[(z.s5 % 256) + charset_offset] << 8;
    (*b2).s6 |= (uint)charset[(z.s6 % 256) + charset_offset] << 8;
    (*b2).s7 |= (uint)charset[(z.s7 % 256) + charset_offset] << 8;
#endif
    if (PasswordLength == 10) {return;}

    z /= (vector_type)256;
#if grt_vector_1 || grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b2).s0 |= (uint)charset[(z.s0 % 256) + charset_offset] << 16;
#endif
#if grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b2).s1 |= (uint)charset[(z.s1 % 256) + charset_offset] << 16;
#endif
#if grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b2).s2 |= (uint)charset[(z.s2 % 256) + charset_offset] << 16;
    (*b2).s3 |= (uint)charset[(z.s3 % 256) + charset_offset] << 16;
#endif
#if grt_vector_8 || grt_vector_16
    (*b2).s4 |= (uint)charset[(z.s4 % 256) + charset_offset] << 16;
    (*b2).s5 |= (uint)charset[(z.s5 % 256) + charset_offset] << 16;
    (*b2).s6 |= (uint)charset[(z.s6 % 256) + charset_offset] << 16;
    (*b2).s7 |= (uint)charset[(z.s7 % 256) + charset_offset] << 16;
#endif
    if (PasswordLength == 11) {return;}

    z /= (vector_type)256;
#if grt_vector_1 || grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b2).s0 |= (uint)charset[(z.s0 % 256) + charset_offset] << 24;
#endif
#if grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b2).s1 |= (uint)charset[(z.s1 % 256) + charset_offset] << 24;
#endif
#if grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b2).s2 |= (uint)charset[(z.s2 % 256) + charset_offset] << 24;
    (*b2).s3 |= (uint)charset[(z.s3 % 256) + charset_offset] << 24;
#endif
#if grt_vector_8 || grt_vector_16
    (*b2).s4 |= (uint)charset[(z.s4 % 256) + charset_offset] << 24;
    (*b2).s5 |= (uint)charset[(z.s5 % 256) + charset_offset] << 24;
    (*b2).s6 |= (uint)charset[(z.s6 % 256) + charset_offset] << 24;
    (*b2).s7 |= (uint)charset[(z.s7 % 256) + charset_offset] << 24;
#endif
    if (PasswordLength == 12) {return;}
}


void searchForHash(uint a, uint b, uint c, uint d, uint vectorPosition,
        LOCAL_MEMORY_TYPE unsigned char *sharedBitmap,
        uint deviceNumberOfHashes, __global unsigned int *deviceHashArray,
        __global   unsigned char *foundPasswordArray, __global  unsigned char *deviceSuccessArray,
        uint b0, uint b1, uint b2, uint b3) {

    uint search_high, search_low, search_index, temp, hash_order_mem, hash_order_a;

#if CPU_DEBUG && 0
    printf("Searching: %c%c%c%c%c%c\n", b0 & 0xff, (b0 >> 8 ) & 0xff,
            (b0 >> 16) & 0xff, (b0 >> 24) & 0xff, b1 & 0xff, (b1 >> 8 ) & 0xff);
#endif

    if ((sharedBitmap[(a & 0x0000ffff) >> 3] >> (a & 0x00000007)) & 0x00000001) {
        //printf("Shared bitmap hit!\n");
        //printf("%08x %08x %08x %08x\n", a, b, c, d);
        
            {
                search_high = deviceNumberOfHashes;
                search_low = 0;
                search_index = 0;
                while (search_low < search_high) {
                    search_index = search_low + (search_high - search_low) / 2;
                    temp = deviceHashArray[4 * search_index];
                    hash_order_mem = (temp & 0xff) << 24 | ((temp >> 8) & 0xff) << 16 | ((temp >> 16) & 0xff) << 8 | ((temp >> 24) & 0xff);
                    hash_order_a = (a & 0xff) << 24 | ((a >> 8) & 0xff) << 16 | ((a >> 16) & 0xff) << 8 | ((a >> 24) & 0xff);
                    if (hash_order_mem < hash_order_a) {
                        search_low = search_index + 1;
                    } else {
                        search_high = search_index;
                    }
                    if ((hash_order_a == hash_order_mem) && (search_low < deviceNumberOfHashes)) {
                        break;
                    }
                }
                if (hash_order_a != hash_order_mem) {
                    goto next;
                }
                while (search_index && (a == deviceHashArray[(search_index - 1) * 4])) {
                    search_index--;
                }
                while ((a == deviceHashArray[search_index * 4])) {
                    {
                        if (b == deviceHashArray[search_index * 4 + 1]) {
                            if (c == deviceHashArray[search_index * 4 + 2]) {
                                if (d == deviceHashArray[search_index * 4 + 3]) {
                                    #ifdef CPU_DEBUG
                                    printf("HASH FOUND!\n");
                                    printf("%08x %08x\n", b0, b1);
                                    #endif
                                    if (PASSWORD_LENGTH >= 1) foundPasswordArray[search_index * MAX_PASSWORD_LENGTH + 0] = (b0 >> 0) & 0xff;
                                    if (PASSWORD_LENGTH >= 2) foundPasswordArray[search_index * MAX_PASSWORD_LENGTH + 1] = (b0 >> 8) & 0xff;
                                    if (PASSWORD_LENGTH >= 3) foundPasswordArray[search_index * MAX_PASSWORD_LENGTH + 2] = (b0 >> 16) & 0xff;
                                    if (PASSWORD_LENGTH >= 4) foundPasswordArray[search_index * MAX_PASSWORD_LENGTH + 3] = (b0 >> 24) & 0xff;
                                    if (PASSWORD_LENGTH >= 5) foundPasswordArray[search_index * MAX_PASSWORD_LENGTH + 4] = (b1 >> 0) & 0xff;
                                    if (PASSWORD_LENGTH >= 6) foundPasswordArray[search_index * MAX_PASSWORD_LENGTH + 5] = (b1 >> 8) & 0xff;
                                    if (PASSWORD_LENGTH >= 7) foundPasswordArray[search_index * MAX_PASSWORD_LENGTH + 6] = (b1 >> 16) & 0xff;
                                    if (PASSWORD_LENGTH >= 8) foundPasswordArray[search_index * MAX_PASSWORD_LENGTH + 7] = (b1 >> 24) & 0xff;
                                    if (PASSWORD_LENGTH >= 9) foundPasswordArray[search_index * MAX_PASSWORD_LENGTH + 8] = (b2 >> 0) & 0xff;
                                    if (PASSWORD_LENGTH >= 10) foundPasswordArray[search_index * MAX_PASSWORD_LENGTH + 9] = (b2 >> 8) & 0xff;
                                    if (PASSWORD_LENGTH >= 11) foundPasswordArray[search_index * MAX_PASSWORD_LENGTH + 10] = (b2 >> 16) & 0xff;
                                    if (PASSWORD_LENGTH >= 12) foundPasswordArray[search_index * MAX_PASSWORD_LENGTH + 11] = (b2 >> 24) & 0xff;
                                    if (PASSWORD_LENGTH >= 13) foundPasswordArray[search_index * MAX_PASSWORD_LENGTH + 12] = (b3 >> 0) & 0xff;
                                    if (PASSWORD_LENGTH >= 14) foundPasswordArray[search_index * MAX_PASSWORD_LENGTH + 13] = (b3 >> 8) & 0xff;
                                    if (PASSWORD_LENGTH >= 15) foundPasswordArray[search_index * MAX_PASSWORD_LENGTH + 14] = (b3 >> 16) & 0xff;
                                    if (PASSWORD_LENGTH >= 16) foundPasswordArray[search_index * MAX_PASSWORD_LENGTH + 15] = (b3 >> 24) & 0xff;
                                    deviceSuccessArray[search_index] = (unsigned char) 1;
                                    
                                }
                            }
                        }
                    }
                    search_index++;
                }
            }
        }
next:
;
}


__kernel __attribute__((vec_type_hint(vector_type))) void RegenerateChainsMD5AMD(
    __constant unsigned char *deviceCharset,/* 0 */
    __private unsigned int deviceCharsetLength,/* 1 */
    __private unsigned int deviceChainLength,/* 2 */
    __private unsigned int deviceTableIndex,/* 3 */
    __private unsigned int deviceNumberOfThreads,/* 4 */
    __private unsigned int deviceNumberOfChainsToRegen, /* 5 */
    __private unsigned int deviceNumberOfHashes, /* 6 */
    __constant  unsigned char *constantBitmap, /* 7 */
    __global   unsigned int *initialPasswordArray, /* 8 */
    __global   unsigned char *foundPasswordArray, /* 9 */
    __global   unsigned int *deviceHashArray, /* 10 */
    __global  unsigned char *deviceSuccessArray, /* 11 */
    __private unsigned int devicePasswordSpaceOffset, /* 12 */
    __private unsigned int deviceStartChainIndex, /* 13 */
    __private unsigned int deviceStepsToRun /* 14 */
) {

#ifndef INTEL_OPENCL_HACK
    // Start the kernel.
    __local unsigned char charset[512];
    __local unsigned char sharedBitmap[8192];
#endif
#if CPU_DEBUG && 0
    printf("\n\n\n");
    printf("Kernel start, global id %d\n", get_global_id(0));
    printf("deviceCharsetLength: %d\n", deviceCharsetLength);
    printf("deviceCharset: %08x ...\n", deviceCharset);
    printf("deviceCharset: %c %c %c %c ...\n", deviceCharset[0], deviceCharset[1], deviceCharset[2], deviceCharset[3]);
    printf("deviceChainLength: %d\n", deviceChainLength);
    printf("deviceTableIndex: %d\n", deviceTableIndex);
    printf("deviceNumberOfThreads: %d\n", deviceNumberOfThreads);
    printf("deviceNumberOfChainsToRegen: %d\n", deviceNumberOfChainsToRegen);
    printf("deviceNumberOfHashes: %d\n", deviceNumberOfHashes);
    printf("devicePasswordSpaceOffset: %d\n", devicePasswordSpaceOffset);
    printf("deviceStartChainIndex: %d\n", deviceStartChainIndex);
    printf("deviceStepsToRun: %d\n", deviceStepsToRun);
    printf("PASSWORD_LENGTH: %d\n", PASSWORD_LENGTH);
#endif


    // Needed variables for generation
    uint CurrentStep, PassCount, password_index, charsetOffset;

    // Hash variables
    vector_type b0,b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15;
    vector_type a,b,c,d;
    vector_type bitmap_index;
#ifndef INTEL_OPENCL_HACK
    // Generic "copy charset to shared memory" function
    copySingleCharsetToShared(charset, deviceCharset);
    //printf("Charset copied... %c %c %c %c ...\n", charset[0], charset[1], charset[2], charset[3]);

    copySingleBitmapToShared(sharedBitmap, constantBitmap);
#endif

    // Figure out which password we are working on.
    password_index = (get_global_id(0) + (devicePasswordSpaceOffset * deviceNumberOfThreads));
#if CPU_DEBUG
    printf("password index: %d\n", password_index);
    printf("startChainIndex: %d\n", deviceStartChainIndex);
#endif
    // Return if this thread is working on something beyond the end of the password space
    if (password_index >= (deviceNumberOfChainsToRegen / VECTOR_WIDTH)) {
#if CPU_DEBUG && 0
        printf("Returning: pass_index > deviceNumberOfChains / 4\n");
#endif
        return;
    }

    b0 = 0x00000000;
    b1 = 0x00000000;
    b2 = 0x00000000;
    b3 = 0x00000000;
    b4 = 0x00000000;
    b5 = 0x00000000;
    b6 = 0x00000000;
    b7 = 0x00000000;
    b8 = 0x00000000;
    b9 = 0x00000000;
    b10 = 0x00000000;
    b11 = 0x00000000;
    b12 = 0x00000000;
    b13 = 0x00000000;
    b14 = 0x00000000;
    b15 = 0x00000000;


    // Load b0/b1 out of memory
    b0 = vload_type(0 * deviceNumberOfChainsToRegen + password_index, initialPasswordArray); // lengths 1-4
#if PASSWORD_LENGTH > 4
    b1 = vload_type((1 * deviceNumberOfChainsToRegen / VECTOR_WIDTH) + password_index, initialPasswordArray); // Len 5-8
#endif
#if PASSWORD_LENGTH > 8
    b2 = vload_type((2 * deviceNumberOfChainsToRegen / VECTOR_WIDTH) + password_index, initialPasswordArray);
#endif
#if PASSWORD_LENGTH > 12
    #error "Password lengths > 12 are not supported!"
#endif

#if CPU_DEBUG
    printf("Initial password s0: %c%c%c%c%c%c\n", b0.s0 & 0xff, (b0.s0 >> 8 ) & 0xff,
        (b0.s0 >> 16) & 0xff, (b0.s0 >> 24) & 0xff, b1.s0 & 0xff, (b1.s0 >> 8 ) & 0xff);
    printf("Initial password s0: %c%c%c%c%c%c\n", b0.s1 & 0xff, (b0.s1 >> 8 ) & 0xff,
        (b0.s1 >> 16) & 0xff, (b0.s1 >> 24) & 0xff, b1.s1 & 0xff, (b1.s1 >> 8 ) & 0xff);
    printf("Initial password s0: %c%c%c%c%c%c\n", b0.s2 & 0xff, (b0.s2 >> 8 ) & 0xff,
        (b0.s2 >> 16) & 0xff, (b0.s2 >> 24) & 0xff, b1.s2 & 0xff, (b1.s2 >> 8 ) & 0xff);
    printf("Initial password s0: %c%c%c%c%c%c\n", b0.s3 & 0xff, (b0.s3 >> 8 ) & 0xff,
        (b0.s3 >> 16) & 0xff, (b0.s3 >> 24) & 0xff, b1.s3 & 0xff, (b1.s3 >> 8 ) & 0xff);
#endif
    charsetOffset = deviceStartChainIndex % deviceCharsetLength;
    
    for (PassCount = 0; PassCount < deviceStepsToRun; PassCount++) {
        CurrentStep = PassCount + deviceStartChainIndex;
#if CPU_DEBUG 
        printf("\nChain %d, step %d\n", password_index, PassCount);
#endif
        padMDHash(PASSWORD_LENGTH, &b0, &b1, &b2, &b3, &b4, &b5, &b6, &b7, &b8, &b9, &b10, &b11, &b12, &b13, &b14, &b15);
        OpenCL_MD5(b0, b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15, &a, &b, &c, &d);
#if CPU_DEBUG
        printf("\n\n\nMD5 result s0... %08x %08x %08x %08x\n", a.s0, b.s0, c.s0, d.s0);
        printf("MD5 result s1... %08x %08x %08x %08x\n", a.s1, b.s1, c.s1, d.s1);
        printf("MD5 result s2... %08x %08x %08x %08x\n", a.s2, b.s2, c.s2, d.s2);
        printf("MD5 result s3... %08x %08x %08x %08x\n", a.s3, b.s3, c.s3, d.s3);
#endif

        // Do all the searching
#if grt_vector_1 || grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
        searchForHash(a.s0, b.s0, c.s0, d.s0, 0, BITMAP_NAME, deviceNumberOfHashes, deviceHashArray, foundPasswordArray, deviceSuccessArray,
                b0.s0, b1.s0, b2.s0, b3.s0);
#endif
#if grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
        searchForHash(a.s1, b.s1, c.s1, d.s1, 1, BITMAP_NAME, deviceNumberOfHashes, deviceHashArray, foundPasswordArray, deviceSuccessArray,
                b0.s1, b1.s1, b2.s1, b3.s1);
#endif
#if grt_vector_4 || grt_vector_8 || grt_vector_16
        searchForHash(a.s2, b.s2, c.s2, d.s2, 2, BITMAP_NAME, deviceNumberOfHashes, deviceHashArray, foundPasswordArray, deviceSuccessArray,
                b0.s2, b1.s2, b2.s2, b3.s2);
        searchForHash(a.s3, b.s3, c.s3, d.s3, 3, BITMAP_NAME, deviceNumberOfHashes, deviceHashArray, foundPasswordArray, deviceSuccessArray,
                b0.s3, b1.s3, b2.s3, b3.s3);
#endif
#if grt_vector_8 || grt_vector_16
        searchForHash(a.s4, b.s4, c.s4, d.s4, 4, BITMAP_NAME, deviceNumberOfHashes, deviceHashArray, foundPasswordArray, deviceSuccessArray,
                b0.s4, b1.s4, b2.s4, b3.s4);
        searchForHash(a.s5, b.s5, c.s5, d.s5, 5, BITMAP_NAME, deviceNumberOfHashes, deviceHashArray, foundPasswordArray, deviceSuccessArray,
                b0.s5, b1.s5, b2.s5, b3.s5);
        searchForHash(a.s6, b.s6, c.s6, d.s6, 6, BITMAP_NAME, deviceNumberOfHashes, deviceHashArray, foundPasswordArray, deviceSuccessArray,
                b0.s6, b1.s6, b2.s6, b3.s6);
        searchForHash(a.s7, b.s7, c.s7, d.s7, 7, BITMAP_NAME, deviceNumberOfHashes, deviceHashArray, foundPasswordArray, deviceSuccessArray,
                b0.s7, b1.s7, b2.s7, b3.s7);
#endif


        reduceSingleCharsetNormal(&b0, &b1, &b2, a, b, c, d, CurrentStep, CHARSET_NAME, charsetOffset, PASSWORD_LENGTH, deviceTableIndex);
#if CPU_DEBUG
        printf("New password s0: %08x %08x\n", b0.s0, b1.s0);
        printf("New password s1: %08x %08x\n", b0.s1, b1.s1);
        printf("New password s2: %08x %08x\n", b0.s2, b1.s2);
        printf("New password s3: %08x %08x\n", b0.s3, b1.s3);
#endif
#if CPU_DEBUG
        printf("New password s0: %c%c%c%c%c%c\n", b0.s0 & 0xff, (b0.s0 >> 8 ) & 0xff,
            (b0.s0 >> 16) & 0xff, (b0.s0 >> 24) & 0xff, b1.s0 & 0xff, (b1.s0 >> 8 ) & 0xff);
        printf("New password s0: %c%c%c%c%c%c\n", b0.s1 & 0xff, (b0.s1 >> 8 ) & 0xff,
            (b0.s1 >> 16) & 0xff, (b0.s1 >> 24) & 0xff, b1.s1 & 0xff, (b1.s1 >> 8 ) & 0xff);
        printf("New password s0: %c%c%c%c%c%c\n", b0.s2 & 0xff, (b0.s2 >> 8 ) & 0xff,
            (b0.s2 >> 16) & 0xff, (b0.s2 >> 24) & 0xff, b1.s2 & 0xff, (b1.s2 >> 8 ) & 0xff);
        printf("New password s0: %c%c%c%c%c%c\n", b0.s3 & 0xff, (b0.s3 >> 8 ) & 0xff,
            (b0.s3 >> 16) & 0xff, (b0.s3 >> 24) & 0xff, b1.s3 & 0xff, (b1.s3 >> 8 ) & 0xff);
#endif
        charsetOffset++;
        if (charsetOffset >= deviceCharsetLength) {
            charsetOffset = 0;
        }
    }
    // Done with the number of steps we need to run

    // If we are done (or have somehow overflowed), store the result
    if (CurrentStep >= (deviceChainLength - 1)) {
#if CPU_DEBUG
        printf("Out of chains, returning.\n");
        printf("final state s0: %08x %08x\n", a.s0, b.s0);
        printf("final state s1: %08x %08x\n", a.s1, b.s1);
        printf("final state s2: %08x %08x\n", a.s2, b.s2);
        printf("final state s3: %08x %08x\n", a.s3, b.s3);
#endif
        return;
    }
    // Else, store the b0/b1 values back to the initial array for the next loop
    else {
#if CPU_DEBUG
        printf("storing state s0: %08x %08x\n", b0.s0, b1.s0);
        printf("storing state s1: %08x %08x\n", b0.s1, b1.s1);
        printf("storing state s2: %08x %08x\n", b0.s2, b1.s2);
        printf("storing state s3: %08x %08x\n", b0.s3, b1.s3);
#endif
        vstore_type(b0, 0 * deviceNumberOfChainsToRegen + password_index, initialPasswordArray);
#if PASSWORD_LENGTH > 4
        vstore_type(b1, (1 * deviceNumberOfChainsToRegen / VECTOR_WIDTH) + password_index, initialPasswordArray);
#endif
#if PASSWORD_LENGTH > 8
        vstore_type(b2, (2 * deviceNumberOfChainsToRegen / VECTOR_WIDTH) + password_index, initialPasswordArray);
#endif
    }

}