// Kernels for MD5 - vectorized for AMD...

// Things we should define in the calling code...
//#define VECTOR_WIDTH 4
#define CPU_DEBUG 1
//#define BITALIGN
//#define NVIDIA_HACKS 1
//#define PASSWORD_LENGTH 6

#define MAX_PASSWORD_LENGTH 32

#if VECTOR_WIDTH == 1
    #define vector_type uint
    #define vload_type vload1
    #define vstore_type vstore1
    #define grt_vector_1 1
#elif VECTOR_WIDTH == 2
    #define vector_type uint2
    #define vload_type vload2
    #define vstore_type vstore2
    #define grt_vector_2 1
#elif VECTOR_WIDTH == 4
    #define vector_type uint4
    #define vload_type vload4
    #define vstore_type vstore4
    #define grt_vector_4 1
#elif VECTOR_WIDTH == 8
ERROR NOT SUPPORTED
    #define vector_type uint8
    #define vload_type vload8
    #define vstore_type vstore8
    #define grt_vector_8 1
#elif VECTOR_WIDTH == 16
ERROR NOT SUPPORTED
    #define vector_type uint16
    #define vload_type vload16
    #define vstore_type vstore16
    #define grt_vector_16 1
#endif


#ifdef CPU_DEBUG
#pragma OPENCL EXTENSION cl_amd_printf : enable
#endif

/* ROTATE_LEFT rotates x left n bits.
 */

#ifdef BITALIGN
#pragma OPENCL EXTENSION cl_amd_media_ops : enable
#define MD4ROTATE_LEFT(x, y) amd_bitalign(x, x, (uint)(32 - y))
#define MD4FF(a,b,c,d,x,s) { (a) = (a)+x+(amd_bytealign((b),(c),(d))); (a) = MD4ROTATE_LEFT((a), (s)); }
#define MD4H(x, y, z) ((x) ^ (y) ^ (z))
#define MD4GG(a,b,c,d,x,s) {(a) = (a) + (vector_type)0x5a827999 + (amd_bytealign((b), ((d) | (c)), ((c) & (d)))) +x  ; (a) = MD4ROTATE_LEFT((a), (s)); }
#elif NVIDIA_HACKS
#define MD4ROTATE_LEFT(x, n) (((x) << (n)) | ((x) >> (32-(n))))
#define MD4F(x, y, z) (((x) & (y)) | ((~x) & (z)))
#define MD4G(x, y, z) (((x) & (y)) | ((x) & (z)) | ((y) & (z)))
#define MD4H(x, y, z) ((x) ^ (y) ^ (z))
#define MD4FF(a, b, c, d, x, s) { \
    (a) += MD4F ((b), (c), (d)) + (x); \
    (a) = MD4ROTATE_LEFT ((a), (s)); \
  }
#define MD4GG(a, b, c, d, x, s) { \
    (a) += MD4G ((b), (c), (d)) + (x) + (vector_type)0x5a827999; \
    (a) = MD4ROTATE_LEFT ((a), (s)); \
  }
#else
#define MD4ROTATE_LEFT(x, y) rotate((vector_type)x, (uint)y)
#define MD4F(x, y, z) (((x) & (y)) | ((~x) & (z)))
#define MD4G(x, y, z) (((x) & (y)) | ((x) & (z)) | ((y) & (z)))
#define MD4H(x, y, z) ((x) ^ (y) ^ (z))
#define MD4FF(a, b, c, d, x, s) { \
    (a) += MD4F ((b), (c), (d)) + (x); \
    (a) = MD4ROTATE_LEFT ((a), (s)); \
  }
#define MD4GG(a, b, c, d, x, s) { \
    (a) += MD4G ((b), (c), (d)) + (x) + (vector_type)0x5a827999; \
    (a) = MD4ROTATE_LEFT ((a), (s)); \
  }
#endif


// Hash defines
/* MD4 Defines as per RFC reference implementation */
/*#define MD4F(x, y, z) (((x) & (y)) | ((~x) & (z)))
#define MD4G(x, y, z) (((x) & (y)) | ((x) & (z)) | ((y) & (z)))
#define MD4H(x, y, z) ((x) ^ (y) ^ (z))
#define MD4ROTATE_LEFT(x, n) (((x) << (n)) | ((x) >> (32-(n))))
*/
/*
#define MD4FF(a, b, c, d, x, s) { \
    (a) += MD4F ((b), (c), (d)) + (x); \
    (a) = MD4ROTATE_LEFT ((a), (s)); \
  }
*/
/*
#define MD4GG(a, b, c, d, x, s) { \
    (a) += MD4G ((b), (c), (d)) + (x) + (vector_type)0x5a827999; \
    (a) = MD4ROTATE_LEFT ((a), (s)); \
  }
*/
#define MD4HH(a, b, c, d, x, s) { \
    (a) += MD4H ((b), (c), (d)) + (x) + (vector_type)0x6ed9eba1; \
    (a) = MD4ROTATE_LEFT ((a), (s)); \
  }
#define MD4S11 3
#define MD4S12 7
#define MD4S13 11
#define MD4S14 19
#define MD4S21 3
#define MD4S22 5
#define MD4S23 9
#define MD4S24 13
#define MD4S31 3
#define MD4S32 9
#define MD4S33 11
#define MD4S34 15
/* End MD4 Defines */






inline void OpenCL_MD4(vector_type b0, vector_type b1, vector_type b2, vector_type b3, vector_type b4, vector_type b5, vector_type b6, vector_type b7,
			   vector_type b8, vector_type b9, vector_type b10, vector_type b11, vector_type b12, vector_type b13, vector_type b14, vector_type b15,
			   vector_type *a, vector_type *b, vector_type *c, vector_type *d) {
  *a = (vector_type)0x67452301;
  *b = (vector_type)0xefcdab89;
  *c = (vector_type)0x98badcfe;
  *d = (vector_type)0x10325476;

  MD4FF (*a, *b, *c, *d, b0, MD4S11); /* 1 */
  MD4FF (*d, *a, *b, *c, b1, MD4S12); /* 2 */
  MD4FF (*c, *d, *a, *b, b2, MD4S13); /* 3 */
  MD4FF (*b, *c, *d, *a, b3, MD4S14); /* 4 */
  MD4FF (*a, *b, *c, *d, b4, MD4S11); /* 5 */
  MD4FF (*d, *a, *b, *c, b5, MD4S12); /* 6 */
  MD4FF (*c, *d, *a, *b, b6, MD4S13); /* 7 */
  MD4FF (*b, *c, *d, *a, b7, MD4S14); /* 8 */
  MD4FF (*a, *b, *c, *d, b8, MD4S11); /* 9 */
  MD4FF (*d, *a, *b, *c, b9, MD4S12); /* 10 */
  MD4FF (*c, *d, *a, *b, b10, MD4S13); /* 11 */
  MD4FF (*b, *c, *d, *a, b11, MD4S14); /* 12 */
  MD4FF (*a, *b, *c, *d, b12, MD4S11); /* 13 */
  MD4FF (*d, *a, *b, *c, b13, MD4S12); /* 14 */
  MD4FF (*c, *d, *a, *b, b14, MD4S13); /* 15 */
  MD4FF (*b, *c, *d, *a, b15, MD4S14); /* 16 */

  /* Round 2 */
  MD4GG (*a, *b, *c, *d, b0, MD4S21); /* 17 */
  MD4GG (*d, *a, *b, *c, b4, MD4S22); /* 18 */
  MD4GG (*c, *d, *a, *b, b8, MD4S23); /* 19 */
  MD4GG (*b, *c, *d, *a, b12, MD4S24); /* 20 */
  MD4GG (*a, *b, *c, *d, b1, MD4S21); /* 21 */
  MD4GG (*d, *a, *b, *c, b5, MD4S22); /* 22 */
  MD4GG (*c, *d, *a, *b, b9, MD4S23); /* 23 */
  MD4GG (*b, *c, *d, *a, b13, MD4S24); /* 24 */
  MD4GG (*a, *b, *c, *d, b2, MD4S21); /* 25 */
  MD4GG (*d, *a, *b, *c, b6, MD4S22); /* 26 */
  MD4GG (*c, *d, *a, *b, b10, MD4S23); /* 27 */
  MD4GG (*b, *c, *d, *a, b14, MD4S24); /* 28 */
  MD4GG (*a, *b, *c, *d, b3, MD4S21); /* 29 */
  MD4GG (*d, *a, *b, *c, b7, MD4S22); /* 30 */
  MD4GG (*c, *d, *a, *b, b11, MD4S23); /* 31 */
  MD4GG (*b, *c, *d, *a, b15, MD4S24); /* 32 */


  /* Round 3 */
  MD4HH (*a, *b, *c, *d, b0, MD4S31); /* 33 */
  MD4HH (*d, *a, *b, *c, b8, MD4S32); /* 34 */
  MD4HH (*c, *d, *a, *b, b4, MD4S33); /* 35 */
  MD4HH (*b, *c, *d, *a, b12, MD4S34); /* 36 */
  MD4HH (*a, *b, *c, *d, b2, MD4S31); /* 37 */
  MD4HH (*d, *a, *b, *c, b10, MD4S32); /* 38 */
  MD4HH (*c, *d, *a, *b, b6, MD4S33); /* 39 */
  MD4HH (*b, *c, *d, *a, b14, MD4S34); /* 40 */
  MD4HH (*a, *b, *c, *d, b1, MD4S31); /* 41 */
  MD4HH (*d, *a, *b, *c, b9, MD4S32); /* 42 */
  MD4HH (*c, *d, *a, *b, b5, MD4S33); /* 43 */
  MD4HH (*b, *c, *d, *a, b13, MD4S34); /* 44 */
  MD4HH (*a, *b, *c, *d, b3, MD4S31); /* 45 */
  MD4HH (*d, *a, *b, *c, b11, MD4S32); /* 46 */
  MD4HH (*c, *d, *a, *b, b7, MD4S33); /* 47 */
  MD4HH (*b, *c, *d, *a, b15, MD4S34); /* 48 */

  // Finally, add initial values, as this is the only pass we make.
  *a += (vector_type)0x67452301;
  *b += (vector_type)0xefcdab89;
  *c += (vector_type)0x98badcfe;
  *d += (vector_type)0x10325476;
}


inline void copySingleCharsetToShared(__local unsigned char *sharedCharset,
        __constant unsigned char *constantCharset) {

    // 32-bit accesses are used to help coalesce memory accesses.
    uint a;

    for (a = 0; a < (512); a++) {
        sharedCharset[a] = constantCharset[a];
    }
}

inline void copySingleBitmapToShared(__local unsigned char *sharedBitmap,
        __constant unsigned char *globalBitmap) {

    // 32-bit accesses are used to help coalesce memory accesses.
    uint a;

    for (a = 0; a < (8192); a++) {
        sharedBitmap[a] = globalBitmap[a];
    }
}


inline void padMDHash(int length,
                vector_type *b0, vector_type *b1, vector_type *b2, vector_type *b3, vector_type *b4, vector_type *b5, vector_type *b6, vector_type *b7,
		vector_type *b8, vector_type *b9, vector_type *b10, vector_type *b11, vector_type *b12, vector_type *b13, vector_type *b14, vector_type *b15) {

  // Set length properly (length in bits)
  *b14 = (vector_type)(length * 8);

  if (length == 0) {
    *b0 |= (vector_type)0x00000080;
    return;
  }
  if (length == 1) {
    *b0 |= (vector_type)0x00008000;
    return;
  }
  if (length == 2) {
    *b0 |= (vector_type)0x00800000;
    return;
  }
  if (length == 3) {
    *b0 |= (vector_type)0x80000000;
    return;
  }
  if (length == 4) {
    *b1 |= (vector_type)0x00000080;
    return;
  }
  if (length == 5) {
    *b1 |= (vector_type)0x00008000;
    return;
  }
  if (length == 6) {
    *b1 |= (vector_type)0x00800000;
    return;
  }
  if (length == 7) {
    *b1 |= (vector_type)0x80000000;
    return;
  }
  if (length == 8) {
    *b2 |= (vector_type)0x00000080;
    return;
  }
  if (length == 9) {
    *b2 |= (vector_type)0x00008000;
    return;
  }
  if (length == 10) {
    *b2 |= (vector_type)0x00800000;
    return;
  }
  if (length == 11) {
    *b2 |= (vector_type)0x80000000;
    return;
  }
  if (length == 12) {
    *b3 |= (vector_type)0x00000080;
    return;
  }
  if (length == 13) {
    *b3 |= (vector_type)0x00008000;
    return;
  }
  if (length == 14) {
    *b3 |= (vector_type)0x00800000;
    return;
  }
  if (length == 15) {
    *b3 |= (vector_type)0x80000000;
    return;
  }
  if (length == 16) {
    *b4 |= (vector_type)0x00000080;
    return;
  }
  if (length == 17) {
    *b4 |= (vector_type)0x00008000;
    return;
  }
  if (length == 18) {
    *b4 |= (vector_type)0x00800000;
    return;
  }
  if (length == 19) {
    *b4 |= (vector_type)0x80000000;
    return;
  }
  if (length == 20) {
    *b5 |= (vector_type)0x00000080;
    return;
  }
  if (length == 21) {
    *b5 |= (vector_type)0x00008000;
    return;
  }
  if (length == 22) {
    *b5 |= (vector_type)0x00800000;
    return;
  }
  if (length == 23) {
    *b5 |= (vector_type)0x80000000;
    return;
  }
}


inline void reduceSingleCharsetNTLM(vector_type *b0, vector_type *b1, vector_type *b2,
        vector_type *b3, vector_type *b4, vector_type *b5,
        vector_type a, vector_type b, vector_type c, vector_type d,
        uint CurrentStep, __local unsigned char *charset, uint charset_offset, int PasswordLength, uint Device_Table_Index) {

    vector_type z;
    // Reduce it
    // First 3
    z = (a+(vector_type)CurrentStep+(vector_type)Device_Table_Index) % (vector_type)(256*256*256);
#if grt_vector_1 || grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b0).s0 = (uint)charset[(z.s0 % 256) + charset_offset];
#endif
#if grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b0).s1 = (uint)charset[(z.s1 % 256) + charset_offset];
#endif
#if grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b0).s2 = (uint)charset[(z.s2 % 256) + charset_offset];
    (*b0).s3 = (uint)charset[(z.s3 % 256) + charset_offset];
#endif
#if grt_vector_8 || grt_vector_16
    (*b0).s4 = (uint)charset[(z.s4 % 256) + charset_offset];
    (*b0).s5 = (uint)charset[(z.s5 % 256) + charset_offset];
    (*b0).s6 = (uint)charset[(z.s6 % 256) + charset_offset];
    (*b0).s7 = (uint)charset[(z.s7 % 256) + charset_offset];
#endif
    if (PasswordLength == 1) {return;}

    z /= (vector_type)256;
#if grt_vector_1 || grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b0).s0 |= (uint)charset[(z.s0 % 256) + charset_offset] << 16;
#endif
#if grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b0).s1 |= (uint)charset[(z.s1 % 256) + charset_offset] << 16;
#endif
#if grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b0).s2 |= (uint)charset[(z.s2 % 256) + charset_offset] << 16;
    (*b0).s3 |= (uint)charset[(z.s3 % 256) + charset_offset] << 16;
#endif
#if grt_vector_8 || grt_vector_16
    (*b0).s4 |= (uint)charset[(z.s4 % 256) + charset_offset] << 16;
    (*b0).s5 |= (uint)charset[(z.s5 % 256) + charset_offset] << 16;
    (*b0).s6 |= (uint)charset[(z.s6 % 256) + charset_offset] << 16;
    (*b0).s7 |= (uint)charset[(z.s7 % 256) + charset_offset] << 16;
#endif
    if (PasswordLength == 2) {return;}

    z /= (vector_type)256;
#if grt_vector_1 || grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b1).s0 = (uint)charset[(z.s0 % 256) + charset_offset];
#endif
#if grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b1).s1 = (uint)charset[(z.s1 % 256) + charset_offset];
#endif
#if grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b1).s2 = (uint)charset[(z.s2 % 256) + charset_offset];
    (*b1).s3 = (uint)charset[(z.s3 % 256) + charset_offset];
#endif
#if grt_vector_8 || grt_vector_16
    (*b1).s4 = (uint)charset[(z.s4 % 256) + charset_offset];
    (*b1).s5 = (uint)charset[(z.s5 % 256) + charset_offset];
    (*b1).s6 = (uint)charset[(z.s6 % 256) + charset_offset];
    (*b1).s7 = (uint)charset[(z.s7 % 256) + charset_offset];
#endif
    if (PasswordLength == 3) {return;}

    // Second 3
    z = (b+(vector_type)CurrentStep+(vector_type)Device_Table_Index) % (vector_type)(256*256*256);
#if grt_vector_1 || grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b1).s0 |= (uint)charset[(z.s0 % 256) + charset_offset] << 16;
    printf("b1.s0: %02x\n", charset[(z.s0 % 256) + charset_offset]);
    printf("charset index: %d\n",(z.s0 % 256) + charset_offset);
#endif
#if grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b1).s1 |= (uint)charset[(z.s1 % 256) + charset_offset] << 16;
#endif
#if grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b1).s2 |= (uint)charset[(z.s2 % 256) + charset_offset] << 16;
    (*b1).s3 |= (uint)charset[(z.s3 % 256) + charset_offset] << 16;
#endif
#if grt_vector_8 || grt_vector_16
    (*b1).s4 |= (uint)charset[(z.s4 % 256) + charset_offset] << 16;
    (*b1).s5 |= (uint)charset[(z.s5 % 256) + charset_offset] << 16;
    (*b1).s6 |= (uint)charset[(z.s6 % 256) + charset_offset] << 16;
    (*b1).s7 |= (uint)charset[(z.s7 % 256) + charset_offset] << 16;
#endif
    if (PasswordLength == 4) {return;}

    z /= (vector_type)256;

#if grt_vector_1 || grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b2).s0 = (uint)charset[(z.s0 % 256) + charset_offset];
#endif
#if grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b2).s1 = (uint)charset[(z.s1 % 256) + charset_offset];
#endif
#if grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b2).s2 = (uint)charset[(z.s2 % 256) + charset_offset];
    (*b2).s3 = (uint)charset[(z.s3 % 256) + charset_offset];
#endif
#if grt_vector_8 || grt_vector_16
    (*b2).s4 = (uint)charset[(z.s4 % 256) + charset_offset];
    (*b2).s5 = (uint)charset[(z.s5 % 256) + charset_offset];
    (*b2).s6 = (uint)charset[(z.s6 % 256) + charset_offset];
    (*b2).s7 = (uint)charset[(z.s7 % 256) + charset_offset];
#endif
    if (PasswordLength == 5) {return;}

    z /= (vector_type)256;
#if grt_vector_1 || grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b2).s0 |= (uint)charset[(z.s0 % 256) + charset_offset] << 16;
#endif
#if grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b2).s1 |= (uint)charset[(z.s1 % 256) + charset_offset] << 16;
#endif
#if grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b2).s2 |= (uint)charset[(z.s2 % 256) + charset_offset] << 16;
    (*b2).s3 |= (uint)charset[(z.s3 % 256) + charset_offset] << 16;
#endif
#if grt_vector_8 || grt_vector_16
    (*b2).s4 |= (uint)charset[(z.s4 % 256) + charset_offset] << 16;
    (*b2).s5 |= (uint)charset[(z.s5 % 256) + charset_offset] << 16;
    (*b2).s6 |= (uint)charset[(z.s6 % 256) + charset_offset] << 16;
    (*b2).s7 |= (uint)charset[(z.s7 % 256) + charset_offset] << 16;
#endif
    if (PasswordLength == 6) {return;}

    // Last 2
    z = (c+(vector_type)CurrentStep+(vector_type)Device_Table_Index) % (vector_type)(256*256*256);
#if grt_vector_1 || grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b3).s0 = (uint)charset[(z.s0 % 256) + charset_offset];
#endif
#if grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b3).s1 = (uint)charset[(z.s1 % 256) + charset_offset];
#endif
#if grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b3).s2 = (uint)charset[(z.s2 % 256) + charset_offset];
    (*b3).s3 = (uint)charset[(z.s3 % 256) + charset_offset];
#endif
#if grt_vector_8 || grt_vector_16
    (*b3).s4 = (uint)charset[(z.s4 % 256) + charset_offset];
    (*b3).s5 = (uint)charset[(z.s5 % 256) + charset_offset];
    (*b3).s6 = (uint)charset[(z.s6 % 256) + charset_offset];
    (*b3).s7 = (uint)charset[(z.s7 % 256) + charset_offset];
#endif
    if (PasswordLength == 7) {return;}

    z /= (vector_type)256;
#if grt_vector_1 || grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b3).s0 |= (uint)charset[(z.s0 % 256) + charset_offset] << 16;
#endif
#if grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b3).s1 |= (uint)charset[(z.s1 % 256) + charset_offset] << 16;
#endif
#if grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b3).s2 |= (uint)charset[(z.s2 % 256) + charset_offset] << 16;
    (*b3).s3 |= (uint)charset[(z.s3 % 256) + charset_offset] << 16;
#endif
#if grt_vector_8 || grt_vector_16
    (*b3).s4 |= (uint)charset[(z.s4 % 256) + charset_offset] << 16;
    (*b3).s5 |= (uint)charset[(z.s5 % 256) + charset_offset] << 16;
    (*b3).s6 |= (uint)charset[(z.s6 % 256) + charset_offset] << 16;
    (*b3).s7 |= (uint)charset[(z.s7 % 256) + charset_offset] << 16;
#endif
    if (PasswordLength == 8) {return;}

    z /= (vector_type)256;
#if grt_vector_1 || grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b4).s0 = (uint)charset[(z.s0 % 256) + charset_offset];
#endif
#if grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b4).s1 = (uint)charset[(z.s1 % 256) + charset_offset];
#endif
#if grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b4).s2 = (uint)charset[(z.s2 % 256) + charset_offset];
    (*b4).s3 = (uint)charset[(z.s3 % 256) + charset_offset];
#endif
#if grt_vector_8 || grt_vector_16
    (*b4).s4 = (uint)charset[(z.s4 % 256) + charset_offset];
    (*b4).s5 = (uint)charset[(z.s5 % 256) + charset_offset];
    (*b4).s6 = (uint)charset[(z.s6 % 256) + charset_offset];
    (*b4).s7 = (uint)charset[(z.s7 % 256) + charset_offset];
#endif
    if (PasswordLength == 9) {return;}

    z = (d+(vector_type)CurrentStep+(vector_type)Device_Table_Index) % (vector_type)(256*256*256);
#if grt_vector_1 || grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b4).s0 |= (uint)charset[(z.s0 % 256) + charset_offset] << 16;
#endif
#if grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b4).s1 |= (uint)charset[(z.s1 % 256) + charset_offset] << 16;
#endif
#if grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b4).s2 |= (uint)charset[(z.s2 % 256) + charset_offset] << 16;
    (*b4).s3 |= (uint)charset[(z.s3 % 256) + charset_offset] << 16;
#endif
#if grt_vector_8 || grt_vector_16
    (*b4).s4 |= (uint)charset[(z.s4 % 256) + charset_offset] << 16;
    (*b4).s5 |= (uint)charset[(z.s5 % 256) + charset_offset] << 16;
    (*b4).s6 |= (uint)charset[(z.s6 % 256) + charset_offset] << 16;
    (*b4).s7 |= (uint)charset[(z.s7 % 256) + charset_offset] << 16;
#endif
    if (PasswordLength == 10) {return;}

    z /= (vector_type)256;
#if grt_vector_1 || grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b5).s0 = (uint)charset[(z.s0 % 256) + charset_offset];
#endif
#if grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b5).s1 = (uint)charset[(z.s1 % 256) + charset_offset];
#endif
#if grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b5).s2 = (uint)charset[(z.s2 % 256) + charset_offset];
    (*b5).s3 = (uint)charset[(z.s3 % 256) + charset_offset];
#endif
#if grt_vector_8 || grt_vector_16
    (*b5).s4 = (uint)charset[(z.s4 % 256) + charset_offset];
    (*b5).s5 = (uint)charset[(z.s5 % 256) + charset_offset];
    (*b5).s6 = (uint)charset[(z.s6 % 256) + charset_offset];
    (*b5).s7 = (uint)charset[(z.s7 % 256) + charset_offset];
#endif
    if (PasswordLength == 11) {return;}

    z /= (vector_type)256;
#if grt_vector_1 || grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b5).s0 |= (uint)charset[(z.s0 % 256) + charset_offset] << 16;
#endif
#if grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b5).s1 |= (uint)charset[(z.s1 % 256) + charset_offset] << 16;
#endif
#if grt_vector_4 || grt_vector_8 || grt_vector_16
    (*b5).s2 |= (uint)charset[(z.s2 % 256) + charset_offset] << 16;
    (*b5).s3 |= (uint)charset[(z.s3 % 256) + charset_offset] << 16;
#endif
#if grt_vector_8 || grt_vector_16
    (*b5).s4 |= (uint)charset[(z.s4 % 256) + charset_offset] << 16;
    (*b5).s5 |= (uint)charset[(z.s5 % 256) + charset_offset] << 16;
    (*b5).s6 |= (uint)charset[(z.s6 % 256) + charset_offset] << 16;
    (*b5).s7 |= (uint)charset[(z.s7 % 256) + charset_offset] << 16;
#endif
    if (PasswordLength == 12) {return;}

}

void searchForHashNTLM(uint a, uint b, uint c, uint d, uint vectorPosition,
        __local unsigned char *sharedBitmap,
        uint deviceNumberOfHashes, __global unsigned int *deviceHashArray,
        __global   unsigned char *foundPasswordArray, __global  unsigned char *deviceSuccessArray,
        uint b0, uint b1, uint b2, uint b3, uint b4, uint b5, uint b6) {

    uint search_high, search_low, search_index, temp, hash_order_mem, hash_order_a;

#if CPU_DEBUG && 0
    printf("Searching: %c%c%c%c%c%c\n", b0 & 0xff, (b0 >> 8 ) & 0xff,
            (b0 >> 16) & 0xff, (b0 >> 24) & 0xff, b1 & 0xff, (b1 >> 8 ) & 0xff);
#endif

    if ((sharedBitmap[(a & 0x0000ffff) >> 3] >> (a & 0x00000007)) & 0x00000001) {
        //printf("Shared bitmap hit!\n");
        //printf("%08x %08x %08x %08x\n", a, b, c, d);

            {
                search_high = deviceNumberOfHashes;
                search_low = 0;
                search_index = 0;
                while (search_low < search_high) {
                    search_index = search_low + (search_high - search_low) / 2;
                    temp = deviceHashArray[4 * search_index];
                    hash_order_mem = (temp & 0xff) << 24 | ((temp >> 8) & 0xff) << 16 | ((temp >> 16) & 0xff) << 8 | ((temp >> 24) & 0xff);
                    hash_order_a = (a & 0xff) << 24 | ((a >> 8) & 0xff) << 16 | ((a >> 16) & 0xff) << 8 | ((a >> 24) & 0xff);
                    if (hash_order_mem < hash_order_a) {
                        search_low = search_index + 1;
                    } else {
                        search_high = search_index;
                    }
                    if ((hash_order_a == hash_order_mem) && (search_low < deviceNumberOfHashes)) {
                        break;
                    }
                }
                if (hash_order_a != hash_order_mem) {
                    goto next;
                }
                while (search_index && (a == deviceHashArray[(search_index - 1) * 4])) {
                    search_index--;
                }
                while ((a == deviceHashArray[search_index * 4])) {
                    {
                        if (b == deviceHashArray[search_index * 4 + 1]) {
                            if (c == deviceHashArray[search_index * 4 + 2]) {
                                if (d == deviceHashArray[search_index * 4 + 3]) {
                                    #ifdef CPU_DEBUG
                                    printf("HASH FOUND!\n");
                                    printf("%08x %08x\n", b0, b1);
                                    #endif
                                    if (PASSWORD_LENGTH >= 1) foundPasswordArray[search_index * MAX_PASSWORD_LENGTH + 0] = (b0 >> 0) & 0xff;
                                    if (PASSWORD_LENGTH >= 2) foundPasswordArray[search_index * MAX_PASSWORD_LENGTH + 1] = (b0 >> 16) & 0xff;
                                    if (PASSWORD_LENGTH >= 3) foundPasswordArray[search_index * MAX_PASSWORD_LENGTH + 2] = (b1 >> 0) & 0xff;
                                    if (PASSWORD_LENGTH >= 4) foundPasswordArray[search_index * MAX_PASSWORD_LENGTH + 3] = (b1 >> 16) & 0xff;
                                    if (PASSWORD_LENGTH >= 5) foundPasswordArray[search_index * MAX_PASSWORD_LENGTH + 4] = (b2 >> 0) & 0xff;
                                    if (PASSWORD_LENGTH >= 6) foundPasswordArray[search_index * MAX_PASSWORD_LENGTH + 5] = (b2 >> 16) & 0xff;
                                    if (PASSWORD_LENGTH >= 7) foundPasswordArray[search_index * MAX_PASSWORD_LENGTH + 6] = (b3 >> 0) & 0xff;
                                    if (PASSWORD_LENGTH >= 8) foundPasswordArray[search_index * MAX_PASSWORD_LENGTH + 7] = (b3 >> 16) & 0xff;
                                    if (PASSWORD_LENGTH >= 9) foundPasswordArray[search_index * MAX_PASSWORD_LENGTH + 8] = (b4 >> 0) & 0xff;
                                    if (PASSWORD_LENGTH >= 10) foundPasswordArray[search_index * MAX_PASSWORD_LENGTH + 9] = (b4 >> 16) & 0xff;
                                    if (PASSWORD_LENGTH >= 11) foundPasswordArray[search_index * MAX_PASSWORD_LENGTH + 10] = (b5 >> 0) & 0xff;
                                    if (PASSWORD_LENGTH >= 12) foundPasswordArray[search_index * MAX_PASSWORD_LENGTH + 11] = (b5 >> 16) & 0xff;
                                    if (PASSWORD_LENGTH >= 13) foundPasswordArray[search_index * MAX_PASSWORD_LENGTH + 12] = (b6 >> 0) & 0xff;
                                    if (PASSWORD_LENGTH >= 14) foundPasswordArray[search_index * MAX_PASSWORD_LENGTH + 13] = (b6 >> 16) & 0xff;
                                    deviceSuccessArray[search_index] = (unsigned char) 1;

                                }
                            }
                        }
                    }
                    search_index++;
                }
            }
        }
next:
;
}


__kernel __attribute__((vec_type_hint(vector_type))) void RegenerateChainsNTLMAMD(
    __constant unsigned char *deviceCharset,/* 0 */
    __private unsigned int deviceCharsetLength,/* 1 */
    __private unsigned int deviceChainLength,/* 2 */
    __private unsigned int deviceTableIndex,/* 3 */
    __private unsigned int deviceNumberOfThreads,/* 4 */
    __private unsigned int deviceNumberOfChainsToRegen, /* 5 */
    __private unsigned int deviceNumberOfHashes, /* 6 */
    __constant  unsigned char *constantBitmap, /* 7 */
    __global   unsigned int *initialPasswordArray, /* 8 */
    __global   unsigned char *foundPasswordArray, /* 9 */
    __global   unsigned int *deviceHashArray, /* 10 */
    __global  unsigned char *deviceSuccessArray, /* 11 */
    __private unsigned int devicePasswordSpaceOffset, /* 12 */
    __private unsigned int deviceStartChainIndex, /* 13 */
    __private unsigned int deviceStepsToRun /* 14 */
) {

    // Start the kernel.
    __local unsigned char charset[512];
    __local unsigned char sharedBitmap[8192];

#if CPU_DEBUG
    printf("\n\n\n");
    printf("Kernel start, global id %d\n", get_global_id(0));
    printf("deviceCharsetLength: %d\n", deviceCharsetLength);
    printf("deviceCharset: %08x ...\n", deviceCharset);
    printf("deviceCharset: %c %c %c %c ...\n", deviceCharset[0], deviceCharset[1], deviceCharset[2], deviceCharset[3]);
    printf("deviceChainLength: %d\n", deviceChainLength);
    printf("deviceTableIndex: %d\n", deviceTableIndex);
    printf("deviceNumberOfThreads: %d\n", deviceNumberOfThreads);
    printf("deviceNumberOfChainsToRegen: %d\n", deviceNumberOfChainsToRegen);
    printf("deviceNumberOfHashes: %d\n", deviceNumberOfHashes);
    printf("devicePasswordSpaceOffset: %d\n", devicePasswordSpaceOffset);
    printf("deviceStartChainIndex: %d\n", deviceStartChainIndex);
    printf("deviceStepsToRun: %d\n", deviceStepsToRun);
    printf("PASSWORD_LENGTH: %d\n", PASSWORD_LENGTH);
#endif


    // Needed variables for generation
    uint CurrentStep, PassCount, password_index, charsetOffset;

    // Hash variables
    vector_type b0,b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15;
    vector_type a,b,c,d;
    vector_type bitmap_index;

    // Generic "copy charset to shared memory" function
    copySingleCharsetToShared(charset, deviceCharset);
    //printf("Charset copied... %c %c %c %c ...\n", charset[0], charset[1], charset[2], charset[3]);

    for (int i = 0; i < 512; i++) {
        printf("Charset %d: %02x %c\n", i, charset[i], charset[i]);
    }

    copySingleBitmapToShared(sharedBitmap, constantBitmap);
    // Figure out which password we are working on.
    password_index = (get_global_id(0) + (devicePasswordSpaceOffset * deviceNumberOfThreads));
#if CPU_DEBUG
    printf("password index: %d\n", password_index);
    printf("startChainIndex: %d\n", deviceStartChainIndex);
#endif
    // Return if this thread is working on something beyond the end of the password space
    if (password_index >= (deviceNumberOfChainsToRegen / VECTOR_WIDTH)) {
#if CPU_DEBUG && 0
        printf("Returning: pass_index > deviceNumberOfChains / 4\n");
#endif
        return;
    }

    b0 = 0x00000000;
    b1 = 0x00000000;
    b2 = 0x00000000;
    b3 = 0x00000000;
    b4 = 0x00000000;
    b5 = 0x00000000;
    b6 = 0x00000000;
    b7 = 0x00000000;
    b8 = 0x00000000;
    b9 = 0x00000000;
    b10 = 0x00000000;
    b11 = 0x00000000;
    b12 = 0x00000000;
    b13 = 0x00000000;
    b14 = 0x00000000;
    b15 = 0x00000000;

    // Load b0/b1 out of memory
    b15 = vload_type(0 * deviceNumberOfChainsToRegen + password_index, initialPasswordArray);
    b0 = (b15 & 0xff) | ((b15 & 0xff00) << 8);
#if PASSWORD_LENGTH > 2
    b1 = ((b15 & 0xff0000) >> 16) | ((b15 & 0xff000000) >> 8);
#endif
#if PASSWORD_LENGTH > 4
    b15 = vload_type((1 * deviceNumberOfChainsToRegen / VECTOR_WIDTH) + password_index, initialPasswordArray);
    b2 = (b15 & 0xff) | ((b15 & 0xff00) << 8);
#endif
#if PASSWORD_LENGTH > 6
    b3 = ((b15 & 0xff0000) >> 16) | ((b15 & 0xff000000) >> 8);
#endif
#if PASSWORD_LENGTH > 8
    b15 = vload_type((2 * deviceNumberOfChainsToRegen / VECTOR_WIDTH) + password_index, initialPasswordArray);
    b4 = (b15 & 0xff) | ((b15 & 0xff00) << 8);
#endif
#if PASSWORD_LENGTH > 10
    b5 = ((b15 & 0xff0000) >> 16) | ((b15 & 0xff000000) >> 8);
#endif
    b15 = (vector_type)0x00000000;

#if CPU_DEBUG
    printf("Initial password s0: %c%c%c%c%c%c\n", b0.s0 & 0xff, (b0.s0 >> 16 ) & 0xff,
        (b1.s0) & 0xff, (b1.s0 >> 16) & 0xff, b2.s0 & 0xff, (b2.s0 >> 16 ) & 0xff);
    printf("Initial password s1: %c%c%c%c%c%c\n", b0.s1 & 0xff, (b0.s1 >> 16 ) & 0xff,
        (b1.s1) & 0xff, (b1.s1 >> 16) & 0xff, b2.s1 & 0xff, (b2.s1 >> 16 ) & 0xff);
    printf("Initial password s2: %c%c%c%c%c%c\n", b0.s2 & 0xff, (b0.s2 >> 16 ) & 0xff,
        (b1.s2) & 0xff, (b1.s2 >> 16) & 0xff, b2.s2 & 0xff, (b2.s2 >> 16 ) & 0xff);
    printf("Initial password s3: %c%c%c%c%c%c\n", b0.s3 & 0xff, (b0.s3 >> 16 ) & 0xff,
        (b1.s3) & 0xff, (b1.s3 >> 16) & 0xff, b2.s3 & 0xff, (b2.s3 >> 16 ) & 0xff);
#endif
    charsetOffset = deviceStartChainIndex % deviceCharsetLength;

    for (PassCount = 0; PassCount < deviceStepsToRun; PassCount++) {
        CurrentStep = PassCount + deviceStartChainIndex;
#if CPU_DEBUG
        printf("\nChain %d, step %d\n", password_index, PassCount);
#endif
        padMDHash(PASSWORD_LENGTH * 2, &b0, &b1, &b2, &b3, &b4, &b5, &b6, &b7, &b8, &b9, &b10, &b11, &b12, &b13, &b14, &b15);
        OpenCL_MD4(b0, b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15, &a, &b, &c, &d);
#if CPU_DEBUG
        printf("\n\n\nNTLM result s0... %08x %08x %08x %08x\n", a.s0, b.s0, c.s0, d.s0);
        printf("NTLM result s1... %08x %08x %08x %08x\n", a.s1, b.s1, c.s1, d.s1);
        printf("NTLM result s2... %08x %08x %08x %08x\n", a.s2, b.s2, c.s2, d.s2);
        printf("NTLM result s3... %08x %08x %08x %08x\n", a.s3, b.s3, c.s3, d.s3);
#endif

        // Do all the searching
#if grt_vector_1 || grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
        searchForHashNTLM(a.s0, b.s0, c.s0, d.s0, 0, sharedBitmap, deviceNumberOfHashes, deviceHashArray, foundPasswordArray, deviceSuccessArray,
                b0.s0, b1.s0, b2.s0, b3.s0, b4.s0, b5.s0, b6.s0);
#endif
#if grt_vector_2 || grt_vector_4 || grt_vector_8 || grt_vector_16
        searchForHashNTLM(a.s1, b.s1, c.s1, d.s1, 1, sharedBitmap, deviceNumberOfHashes, deviceHashArray, foundPasswordArray, deviceSuccessArray,
                b0.s1, b1.s1, b2.s1, b3.s1, b4.s1, b5.s1, b6.s1);
#endif
#if grt_vector_4 || grt_vector_8 || grt_vector_16
        searchForHashNTLM(a.s2, b.s2, c.s2, d.s2, 2, sharedBitmap, deviceNumberOfHashes, deviceHashArray, foundPasswordArray, deviceSuccessArray,
                b0.s2, b1.s2, b2.s2, b3.s2, b4.s2, b5.s2, b6.s2);
        searchForHashNTLM(a.s3, b.s3, c.s3, d.s3, 3, sharedBitmap, deviceNumberOfHashes, deviceHashArray, foundPasswordArray, deviceSuccessArray,
                b0.s3, b1.s3, b2.s3, b3.s3, b4.s3, b5.s3, b6.s3);
#endif
#if grt_vector_8 || grt_vector_16
        searchForHashNTLM(a.s4, b.s4, c.s4, d.s4, 4, sharedBitmap, deviceNumberOfHashes, deviceHashArray, foundPasswordArray, deviceSuccessArray,
                b0.s4, b1.s4, b2.s4, b3.s4, b4.s4, b5.s4, b6.s4);
        searchForHashNTLM(a.s5, b.s5, c.s5, d.s5, 5, sharedBitmap, deviceNumberOfHashes, deviceHashArray, foundPasswordArray, deviceSuccessArray,
                b0.s5, b1.s5, b2.s5, b3.s5, b4.s5, b5.s5, b6.s5);
        searchForHashNTLM(a.s6, b.s6, c.s6, d.s6, 6, sharedBitmap, deviceNumberOfHashes, deviceHashArray, foundPasswordArray, deviceSuccessArray,
                b0.s6, b1.s6, b2.s6, b3.s6, b4.s6, b5.s6, b6.s6);
        searchForHashNTLM(a.s7, b.s7, c.s7, d.s7, 7, sharedBitmap, deviceNumberOfHashes, deviceHashArray, foundPasswordArray, deviceSuccessArray,
                b0.s7, b1.s7, b2.s7, b3.s7, b4.s7, b5.s7, b6.s7);
#endif


        reduceSingleCharsetNTLM(&b0, &b1, &b2, &b3, &b4, &b5, a, b, c, d, CurrentStep, charset, charsetOffset, PASSWORD_LENGTH, deviceTableIndex);
#if CPU_DEBUG
        printf("New password s0: %08x %08x\n", b0.s0, b1.s0);
        printf("New password s1: %08x %08x\n", b0.s1, b1.s1);
        printf("New password s2: %08x %08x\n", b0.s2, b1.s2);
        printf("New password s3: %08x %08x\n", b0.s3, b1.s3);
#endif
#if CPU_DEBUG
    printf("New password s0: %c%c%c%c%c%c\n", b0.s0 & 0xff, (b0.s0 >> 16 ) & 0xff,
        (b1.s0) & 0xff, (b1.s0 >> 16) & 0xff, b2.s0 & 0xff, (b2.s0 >> 16 ) & 0xff);
    printf("New password s1: %c%c%c%c%c%c\n", b0.s1 & 0xff, (b0.s1 >> 16 ) & 0xff,
        (b1.s1) & 0xff, (b1.s1 >> 16) & 0xff, b2.s1 & 0xff, (b2.s1 >> 16 ) & 0xff);
    printf("New password s2: %c%c%c%c%c%c\n", b0.s2 & 0xff, (b0.s2 >> 16 ) & 0xff,
        (b1.s2) & 0xff, (b1.s2 >> 16) & 0xff, b2.s2 & 0xff, (b2.s2 >> 16 ) & 0xff);
    printf("New password s3: %c%c%c%c%c%c\n", b0.s3 & 0xff, (b0.s3 >> 16 ) & 0xff,
        (b1.s3) & 0xff, (b1.s3 >> 16) & 0xff, b2.s3 & 0xff, (b2.s3 >> 16 ) & 0xff);
#endif
        charsetOffset++;
        if (charsetOffset >= deviceCharsetLength) {
            charsetOffset = 0;
        }
#if CPU_DEBUG
        printf("charsetOffset: %d\n", charsetOffset);
#endif

    }
    // Done with the number of steps we need to run

    // If we are done (or have somehow overflowed), store the result
    if (CurrentStep >= (deviceChainLength - 1)) {
#if CPU_DEBUG
        printf("Out of chains, returning.\n");
        printf("final state s0: %08x %08x\n", a.s0, b.s0);
        printf("final state s1: %08x %08x\n", a.s1, b.s1);
        printf("final state s2: %08x %08x\n", a.s2, b.s2);
        printf("final state s3: %08x %08x\n", a.s3, b.s3);
#endif
        return;
    }
    // Else, store the b0/b1 values back to the initial array for the next loop
    else {
#if CPU_DEBUG
        printf("storing state s0: %08x %08x\n", b0.s0, b1.s0);
        printf("storing state s1: %08x %08x\n", b0.s1, b1.s1);
        printf("storing state s2: %08x %08x\n", b0.s2, b1.s2);
        printf("storing state s3: %08x %08x\n", b0.s3, b1.s3);
#endif

        b15 =  (b0 & 0xff) | ((b0 & 0xff0000) >> 8);
#if PASSWORD_LENGTH > 2
        b15 |= (b1 & 0xff) << 16 | ((b1 & 0xff0000) << 8);
#endif
#if CPU_DEBUG
        printf("storing state b15: %08x %08x %08x %08x\n", b15.s0, b15.s1, b15.s2, b15.s3);
#endif
        // Need to store this for all cases.
        vstore_type(b15, 0 * deviceNumberOfChainsToRegen + password_index, initialPasswordArray);

#if PASSWORD_LENGTH > 4
        b15 =  (b2 & 0xff) | ((b2 & 0xff0000) >> 8);
    #if PASSWORD_LENGTH > 6
        b15 |= (b3 & 0xff) << 16 | ((b3 & 0xff0000) << 8);

    #endif
        vstore_type(b15, (1 * deviceNumberOfChainsToRegen / VECTOR_WIDTH) + password_index, initialPasswordArray);
#endif

#if PASSWORD_LENGTH > 8
        b15 =  (b4 & 0xff) | ((b4 & 0xff0000) >> 8);
    #if PASSWORD_LENGTH > 10
        b15 |= (b5 & 0xff) << 16 | ((b5 & 0xff0000) << 8);

    #endif
        vstore_type(b15, (2 * deviceNumberOfChainsToRegen / VECTOR_WIDTH) + password_index, initialPasswordArray);
#endif

    }

}